"""Container class for list of Pyscal objects"""

import warnings
from pathlib import Path
from typing import List, Optional, Type, Union

import pandas as pd

from pyscal import (
    GasOil,
    GasWater,
    SCALrecommendation,
    WaterOil,
    WaterOilGas,
    getLogger_pyscal,
)

PYSCAL_OBJECTS = [WaterOil, GasOil, GasWater, WaterOilGas, SCALrecommendation]

PyscalObjects = Union[WaterOil, GasOil, GasWater, WaterOilGas, SCALrecommendation]

logger = getLogger_pyscal(__name__)

warnings.filterwarnings("default", category=DeprecationWarning, module="pyscal")


class PyscalList:
    """Container class for a list of WaterOilGas objects.

    Essentially this is a list of objects of equal type, and all
    being pyscal objects WaterOil, GasOil, WaterOilGas or SCALrecommendation

    It is possible to ask this list class for SWOF++ printouts,
    and it will call SWOF on each element succesively.

    Args:
        pyscal_list (list): List of objects if already ready. Can be empty or None.
    """

    def __init__(self, pyscal_list: Optional[List[PyscalObjects]] = None):
        self.pyscaltype: Optional[Type] = None
        self.pyscal_list: List[PyscalObjects] = []
        if isinstance(pyscal_list, list):
            for pyscal_obj in pyscal_list:
                self.append(pyscal_obj)
        if isinstance(pyscal_list, PyscalList):
            for idx in range(len(pyscal_list)):
                self.append(pyscal_list[idx + 1])

    def append(self, pyscal_obj: Optional[PyscalObjects]) -> None:
        """Append a pyscal object to the list

        Args:
            pyscal_obj

        Raises:
            ValueError
                If the type of the incoming object does not
                match existing objects in the list
        """
        if pyscal_obj is None:
            return
        if isinstance(pyscal_obj, list):
            # Recursion
            for pyscal_obj_sub in pyscal_obj:
                self.append(pyscal_obj_sub)
            return
        if not isinstance(pyscal_obj, tuple(PYSCAL_OBJECTS)):
            raise ValueError("Not a pyscal object: " + str(pyscal_obj))
        if not self.pyscaltype:
            self.pyscaltype = type(pyscal_obj)
            # Beware, this list can be of type WaterOilGas, with
            # WaterOilGas objects where gasoil is None, effectively
            # making that object a WaterOil object.
        if not isinstance(pyscal_obj, self.pyscaltype):
            raise ValueError(
                f"Trying to add {type(pyscal_obj)} to list "
                f"of {self.pyscaltype} objects."
            )
        self.pyscal_list.append(pyscal_obj)

    def df(self) -> pd.DataFrame:
        """Dump dataframes of generated relperm data

        Column names are compatible with ecl2df.satfunc. Always uppercase
        and capillary pressure is PCOW or PCOG (wateroil vs gasoil)

        If the PyscalList contains SCALrecommendations, the CASE column
        will contain the strings 'pess', 'base' and 'opt' (independent of
        any alias name potentially used in an input xlsx/csv)
        """
        # Names of dataframe columns in wateroil/gasoil.table:
        wateroil_pyscal_cols = {"SW", "KRW", "KROW", "PC"}
        gasoil_pyscal_cols = {"SG", "KRG", "KROG", "PC"}

        # Renamers applied to the returned dataframe:
        gasoil_col_renamer = {"SG": "SG", "KRG": "KRG", "KROG": "KROG", "PC": "PCOG"}
        wateroil_col_renamer = {"SW": "SW", "KRW": "KRW", "KROW": "KROW", "PC": "PCOW"}

        # Sort order for rows in returned dataframe:
        sort_candidates = ["SATNUM", "CASE", "KEYWORD", "SW", "SG", "SL"]

        df_list = []
        if self.pyscaltype == WaterOilGas:
            for satnum, wateroilgas in enumerate(self.pyscal_list):
                assert isinstance(wateroilgas, WaterOilGas)
                assert wateroilgas.wateroil is not None
                assert wateroilgas.gasoil is not None
                wateroil_cols = list(
                    set(wateroilgas.wateroil.table.columns).intersection(
                        wateroil_pyscal_cols
                    )
                )
                gasoil_cols = list(
                    set(wateroilgas.gasoil.table.columns).intersection(
                        gasoil_pyscal_cols
                    )
                )
                df_list.append(
                    wateroilgas.gasoil.table[gasoil_cols]
                    .assign(SATNUM=satnum + 1)
                    .rename(gasoil_col_renamer, axis="columns")
                )
                df_list.append(
                    wateroilgas.wateroil.table[wateroil_cols]
                    .assign(SATNUM=satnum + 1)
                    .rename(wateroil_col_renamer, axis="columns")
                )
        elif self.pyscaltype == SCALrecommendation:
            for satnum, scalrec in enumerate(self.pyscal_list):
                assert isinstance(scalrec, SCALrecommendation)
                assert scalrec.low is not None
                assert scalrec.base is not None
                assert scalrec.high is not None
                assert scalrec.low.wateroil is not None
                assert scalrec.low.gasoil is not None
                assert scalrec.base.wateroil is not None
                assert scalrec.base.gasoil is not None
                assert scalrec.high.wateroil is not None
                assert scalrec.high.gasoil is not None
                gasoil_cols = list(
                    set(scalrec.base.gasoil.table.columns).intersection(
                        gasoil_pyscal_cols
                    )
                )
                wateroil_cols = list(
                    set(scalrec.base.wateroil.table.columns).intersection(
                        wateroil_pyscal_cols
                    )
                )
                df_list.append(
                    scalrec.low.gasoil.table[gasoil_cols]
                    .assign(SATNUM=satnum + 1, CASE="pess")
                    .rename(gasoil_col_renamer, axis="columns")
                )
                df_list.append(
                    scalrec.base.gasoil.table[gasoil_cols]
                    .assign(SATNUM=satnum + 1, CASE="base")
                    .rename(gasoil_col_renamer, axis="columns")
                )
                df_list.append(
                    scalrec.high.gasoil.table[gasoil_cols]
                    .assign(SATNUM=satnum + 1, CASE="opt")
                    .rename(gasoil_col_renamer, axis="columns")
                )

                df_list.append(
                    scalrec.low.wateroil.table[wateroil_cols]
                    .assign(SATNUM=satnum + 1, CASE="pess")
                    .rename(wateroil_col_renamer, axis="columns")
                )
                df_list.append(
                    scalrec.base.wateroil.table[wateroil_cols]
                    .assign(SATNUM=satnum + 1, CASE="base")
                    .rename(wateroil_col_renamer, axis="columns")
                )
                df_list.append(
                    scalrec.high.wateroil.table[wateroil_cols]
                    .assign(SATNUM=satnum + 1, CASE="opt")
                    .rename(wateroil_col_renamer, axis="columns")
                )
        elif self.pyscaltype == WaterOil:
            for satnum, wateroil in enumerate(self.pyscal_list):
                assert isinstance(wateroil, WaterOil)
                assert wateroil is not None
                wateroil_cols = list(
                    set(wateroil.table.columns).intersection(wateroil_pyscal_cols)
                )
                df_list.append(
                    wateroil.table[wateroil_cols]
                    .assign(SATNUM=satnum + 1)
                    .rename(wateroil_col_renamer, axis="columns")
                )
        elif self.pyscaltype == GasOil:
            for satnum, gasoil in enumerate(self.pyscal_list):
                assert isinstance(gasoil, GasOil)
                assert gasoil is not None
                gasoil_cols = list(
                    set(gasoil.table.columns).intersection(gasoil_pyscal_cols)
                )
                df_list.append(
                    gasoil.table[gasoil_cols]
                    .assign(SATNUM=satnum + 1)
                    .rename(gasoil_col_renamer, axis="columns")
                )
        dframe = pd.concat(df_list, sort=False, ignore_index=True)
        sort_rows_on = [colname for colname in sort_candidates if colname in dframe]
        if sort_rows_on:
            dframe = dframe.sort_values(sort_rows_on)
        return dframe

    def relevant_keywords(self, family: int = 1, slgof: bool = False) -> List[str]:
        """Construct a list of relevant Eclipse keywords for the data in this
        Pyscallist object. This depends on the Pyscaltype, and which family is
        requested"""

        if family not in [1, 2]:
            raise ValueError("Family must be either 1 or 2")

        if self.pyscaltype == WaterOilGas:
            # WaterOilGas can be of type WaterOil or GasOil when it emerges
            # from a SCAL recommendation, signified by None-ness of attributes
            if family == 2:
                return ["SWFN", "SGFN", "SOF3"]
            if self.pyscal_list[0].gasoil is None:  # type: ignore
                return ["SWOF"]
            if self.pyscal_list[0].wateroil is None:  # type: ignore
                return ["SGOF"]
            if not slgof:
                return ["SWOF", "SGOF"]
            return ["SWOF", "SLGOF"]

        if self.pyscaltype == WaterOil:
            if family == 2:
                raise ValueError("Family 2 only supported for WaterOilGas and GasWater")
            return ["SWOF"]

        if self.pyscaltype == GasOil:
            if family == 2:
                raise ValueError("Family 2 only supported for WaterOilGas and GasWater")
            if slgof:
                raise ValueError("SLGOF not meaningful for GasOil")
            return ["SGOF"]

        assert self.pyscaltype == GasWater
        if family == 2:
            return ["SWFN", "SGFN"]
        raise ValueError("Family 1 output not possible for GasWater")

    def build_eclipse_data(self, family: int = 1, slgof: bool = False) -> str:
        """Construct Eclipse keywords and data for relative permeability
        properties of family 1 or 2 type.

        Args:
            slgof: Set to true of SLGOF is wanted instead of SGOF. Only applicable
            if family is 1.
        """
        if family not in [1, 2]:
            raise ValueError("Family must be either 1 or 2")
        if len(self) == 0:
            return ""
        if self.pyscaltype == SCALrecommendation:
            raise TypeError(
                "You need to interpolate before you can dump a SCAL recommendation"
            )
        if family == 2 and slgof is True:
            raise ValueError("SLGOF not meaningful for family 2")
        keywords = self.relevant_keywords(family=family, slgof=slgof)
        logger.info(
            "Keywords %s (family %d) for %d SATNUMs generated",
            ", ".join(keywords),
            family,
            len(self),
        )
        return "\n".join([getattr(self, keyword)() for keyword in keywords])

    def dump_family_1(self, filename: Optional[str] = None, slgof: bool = False) -> str:
        """Dumps family 1 Eclipse saturation tables to one
        filename. This means SWOF + SGOF (SGOF only if relevant)

        This function is deprecated. Use build_eclipse_data() and write to
        disk in calling code.

        Args:
            filename: Filename for the output to be given to Eclipse 100
            slgof: Set to true of SLGOF is wanted instead of SGOF
        """
        warnings.warn("dump_family_1() is deprecated", DeprecationWarning)
        string = self.build_eclipse_data(family=1, slgof=slgof)
        if filename is not None:
            if not Path(filename).parent.exists():
                raise IOError(f"Output directory not found '{Path(filename).parent}'")
            Path(filename).write_text(string, encoding="utf-8")
        return string

    def dump_family_2(self, filename: Optional[str] = None) -> str:
        """Dumps family 2 Eclipse saturation tables to one
        filename. This means SWFN + SGFN + SOF3 (SOF3 only for WaterOilGas)

        Relevant for WaterOilGas and GasWater.

        Args:
            filename (str): Filename for the output to be given to Eclipse 100
        """
        warnings.warn("dump_family_2() is deprecated", DeprecationWarning)
        string = self.build_eclipse_data(family=2, slgof=False)
        if filename is not None:
            if not Path(filename).parent.exists():
                raise IOError(f"Output directory not found '{Path(filename).parent}'")

            Path(filename).write_text(string, encoding="utf-8")
        return string

    def interpolate(
        self,
        int_params_wo: Union[float, int, List[float]],
        int_params_go: Optional[Union[float, int, List[Optional[float]]]] = None,
        h: Optional[float] = None,
    ) -> "PyscalList":
        """This function will interpolate each SCALrecommendation
        object to the chosen parameters

        This only works on lists of SCALrecommendation objects

        Args:
            int_params_wo: Interpolation parameters for wateroil, or for
                both. If list, separate parameter for each SATNUM. All
                numbers between -1 and 1 (inclusive).
            int_params_go: If specified, will be used for GasOil interpolation.
            h: Saturation step-length

        Returns:
            PyscalList of type WaterOilGas, with the same length.
        """

        if self.pyscaltype != SCALrecommendation:
            raise TypeError(
                "Can only interpolate PyscalList of type SCALrecommendation"
            )
        if isinstance(int_params_wo, (float, int)):
            int_params_wo = [int_params_wo] * len(self)
        elif isinstance(int_params_wo, list) and len(int_params_wo) == 1:
            int_params_wo = int_params_wo * len(self)
        if int_params_go is None or isinstance(int_params_go, (float, int)):
            int_params_go = [int_params_go] * len(self)
        elif isinstance(int_params_go, list) and len(int_params_go) == 1:
            int_params_go = int_params_go * len(self)

        if 1 < len(int_params_wo) < len(self):
            raise ValueError(
                f"Too few interpolation parameters given for WaterOil {int_params_wo}"
            )
        if len(int_params_wo) > len(self):
            raise ValueError(
                f"Too many interpolation parameters given for WaterOil {int_params_wo}",
            )
        if 1 < len(int_params_go) < len(self):
            raise ValueError(
                f"Too few interpolation parameters given for GasOil {int_params_go}"
            )
        if len(int_params_go) > len(self):
            raise ValueError(
                f"Too many interpolation parameters given for GasOil {int_params_go}"
            )
        wog_list: PyscalList = PyscalList()
        for satnum, scalrec in enumerate(self.pyscal_list):
            assert isinstance(scalrec, SCALrecommendation)
            wog_list.append(
                scalrec.interpolate(int_params_wo[satnum], int_params_go[satnum], h=h)
            )
        return wog_list

    def _make_ecl_output(
        self,
        keyword: str,
        write_to_filename: Optional[str] = None,  # Deprecated
    ) -> str:
        """Internal helper function for constructing Eclipse include file strings
        for individual keywords.

        build_eclipse_data() will use this function.
        """
        if self.pyscaltype == SCALrecommendation:
            raise TypeError(
                "You need to interpolate before you can dump a SCAL recommendation"
            )
        first_obj = self.pyscal_list[0]
        outputter = getattr(first_obj, keyword)
        string = outputter(header=True)
        if len(self.pyscal_list) > 1:
            for pyscal_obj in self.pyscal_list[1:]:
                outputter = getattr(pyscal_obj, keyword)
                string += outputter(header=False)
        if write_to_filename:
            warnings.warn(
                "Writing to files in pyscallist is deprecated", DeprecationWarning
            )
            Path(write_to_filename).parent.mkdir(parents=True, exist_ok=True)
            Path(write_to_filename).write_text(string, encoding="utf-8")
        return string

    def SWOF(self, write_to_filename: Optional[str] = None) -> str:
        """Build SWOF string"""
        # _make_ecl_output() will warn about non-None filename being deprecated
        return self._make_ecl_output("SWOF", write_to_filename)

    def SGOF(self, write_to_filename: Optional[str] = None) -> str:
        """Build SGOF string"""
        return self._make_ecl_output("SGOF", write_to_filename)

    def SLGOF(self, write_to_filename: Optional[str] = None) -> str:
        """Build SLGOF string"""
        return self._make_ecl_output("SLGOF", write_to_filename)

    def SGFN(self, write_to_filename: Optional[str] = None) -> str:
        """Build SGFN string"""
        return self._make_ecl_output("SGFN", write_to_filename)

    def SWFN(self, write_to_filename: Optional[str] = None) -> str:
        """Build SWFN string"""
        return self._make_ecl_output("SWFN", write_to_filename)

    def SOF3(self, write_to_filename: Optional[str] = None) -> str:
        """Build SOF3 string"""
        return self._make_ecl_output("SOF3", write_to_filename)

    def __len__(self) -> int:
        """Return the count of Pyscal objects in the list"""
        return len(self.pyscal_list)

    def __getitem__(self, satnum_idx) -> PyscalObjects:
        """Get a specific List member.

        The indexing starts at 1, not zero, similar
        to how SATNUMs are indexed.

        Args:
            satnum_idx (int): Index for wanted SATNUM. Starts at 1

        Returns:
            WaterOilGas, GasOil, WaterOil or SCALrecommendation, depending
                on self.pyscaltype.
        """
        if satnum_idx < 1:
            e_msg = "SATNUM must be 1 or higher"
            logger.error(e_msg)
            raise IndexError(e_msg)
        if satnum_idx > self.__len__():
            e_msg = "SATNUM index out of range, length is " + str(self.__len__())
            logger.error(e_msg)
            raise IndexError(e_msg)
        return self.pyscal_list[satnum_idx - 1]



"""Utility functions for computations on relative permeability curves"""

import logging
from typing import Optional

import numpy as np
import pandas as pd

from ..constants import EPSILON as epsilon
from ..constants import SWINTEGERS

logger = logging.getLogger(__name__)


def truncate_zeroness(
    value: float, zeronesslimit: float = 10 / SWINTEGERS, name: str = "", log=True
) -> float:
    """Check a value for closeness to zero, and return as zero if below
    a given limit, if not return the value"""
    if value < zeronesslimit:
        if log and name and not np.isclose(value, 0.0):
            logger.warning(f"{name}={value} was close to zero, truncated to 0")
        return 0.0
    return value


def crosspoint(dframe: pd.DataFrame, satcol: str, kr1col: str, kr2col: str) -> float:
    """Locate the saturation value (crosspoint) where kr1col == kr2col

    Args:
        dframe: Dataframe with at least three columns
        satcol: Column name for the saturation column
        kr1col: Column name for first relperm column
        kr2col: Column name for second column

    Returns:
        The saturation value (interpolated) where kr1col == kr2col, when krXcol
        is linearly interpolated as a function of the saturation values. In case
        of errors, the function will return the value -1
    """
    if len(dframe) < 2:
        return -1

    cross_dframe = pd.DataFrame(dframe[[satcol, kr1col, kr2col]])

    if cross_dframe.isna().any().any():
        logger.error("nan in input to crosspoint()")
        logger.debug(str(cross_dframe))
        return -1

    cross_dframe.loc[:, "krdiff"] = cross_dframe[kr1col] - cross_dframe[kr2col]

    # Add a zero value for the difference column, and interpolate
    # the saturation column to the zero value
    zerodf = pd.DataFrame(index=[len(cross_dframe)], data={"krdiff": 0.0})
    cross_dframe = pd.concat([cross_dframe, zerodf], sort=True).set_index("krdiff")

    cross_dframe = cross_dframe[~cross_dframe.index.duplicated(keep="first")]

    if cross_dframe.shape[0] > 2:
        cross_dframe = cross_dframe.interpolate(method="slinear")

    if cross_dframe.isna().any().any():
        logger.error("Could not compute crosspoint)")
        logger.debug(str(cross_dframe))
        return -1

    return cross_dframe[np.isclose(cross_dframe.index, 0.0)][satcol].to_numpy()[0]


def estimate_diffjumppoint(
    table: pd.DataFrame,
    xcol: Optional[str] = None,
    ycol: Optional[str] = None,
    side: str = "right",
) -> float:
    """Estimate the point where the y-data jumps from being linear
    in x to being nonlinear, or where it shift from one linear domain
    to another (for a piecewise linear function)

    If xcol is sw, and ycol is krw, and side is 'right', this
    will typically estimate sorw for you. If side is 'left' it will
    give you swcr.

    Args:
        table: A Dataframe with x and y data
        xcol: The name of the column in table containing x-data. If
            None (default) the first column in table will be used.
        ycol: The name of the column in table containing y-data.
            If None (default) the second column in the table will be used.
        side: Must be 'left' or 'right'. Decides whether to look from
            the right side of the x-interval or from the left side for the
            linear domain.
    Returns:
        The x value where the start-linear domain ends.
    """

    if not xcol:
        xcol = table.columns[0]
    if not ycol:
        ycol = table.columns[1]
    assert isinstance(ycol, str)
    assert isinstance(xcol, str)
    if not side:
        raise ValueError("side cannot be None, use left or right")
    side = side.lower()
    assert side in ["left", "right"]

    # Compute the derivative:
    table["_deriv"] = table[ycol].diff() / table[xcol].diff()
    # The first becomes NaN, extrapolate from the second row:
    table.loc[0, "_deriv"] = table["_deriv"].iloc[1]

    # Pick the derivative at the first or last segment:
    iloc = {"left": 0, "right": -1}
    lin_a = table["_deriv"].iloc[iloc[side]]

    # Make a linear extrapolation from the last segment, starting at max x
    table["_linear"] = (table[xcol] - table[xcol].iloc[iloc[side]]) * lin_a + table[
        ycol
    ].iloc[iloc[side]]
    assert table["_linear"].to_numpy()[iloc[side]] == table[ycol].to_numpy()[iloc[side]]

    # Compute how much krw deviates from the linear krw:
    table["_lindev"] = (table[ycol] - table["_linear"]).abs()

    # Use the cumulative sum to determine the onset of non-zero deviation
    # starting from sw=1:
    table["_lindevcumsum"] = table["_lindev"].cumsum()

    if side == "right":
        maxcumsum = table["_lindevcumsum"].max()
        linearpart = table[(table["_lindevcumsum"] - maxcumsum).abs() < epsilon]
        return linearpart.iloc[1][xcol]

    linearpart = table[(table["_lindevcumsum"] < epsilon)]
    if len(linearpart) == 1:
        linearpart = table[(table["_lindevcumsum"].shift(1) < epsilon)]
    return linearpart.iloc[-1][xcol]



"""Support functions for capillary pressure"""

import math

from pyscal.constants import MAX_EXPONENT_PC


def simple_J(
    # sw: Union[float, Iterable[float]],
    sw: float,
    a: float,
    b: float,
    poro_ref: float,
    perm_ref: float,
    drho: float,
    g: float = 9.81,
) -> float:  # Union[float, Iterable[float]]:
    r"""Calculate capillary pressure with bar as unit

    RMS version:

    .. math::

        J = a S_w^b

    *J* is not dimensionless in this equation.

    This is identical to the also seen formula

    .. math::

       J = 10^{b \log(S_w) + \log(a)}


    Args:
        sw: float, water saturation value to be used.
            Normalize when needed.
        a: a coefficient
        b: b coefficient
        poro_ref: Reference porosity for scaling to Pc, between 0 and 1
        perm_ref: Reference permeability for scaling to Pc, in milliDarcy
        drho: Density difference between water and oil, in SI units kg/m³.
            Default value is 300
        g: Gravitational acceleration, in SI units m/s²,
            default value is 9.81

    Returns:
        capillary pressure, same type as swnpc input argument.
    """
    assert g >= 0
    assert b < MAX_EXPONENT_PC
    assert b > -MAX_EXPONENT_PC
    assert 0.0 <= poro_ref <= 1.0
    assert perm_ref > 0.0

    J = _sw_to_simpleJ(sw, a, b)
    height = _simpleJ_to_height(J, poro_ref, perm_ref)
    return _height_to_pc(height, drho, g)


def _height_to_pc(height: float, drho: float, g: float) -> float:
    """From height above free water level, multiplication
    with density difference and gravity gives capillary pressure.

    Args:
        height: Height above free water level in meters.
        drho: density difference in g/cc
        g: gravitational acceleration, in m/s²

    Returns:
        float: capillary pressure at given height, in bars.
    """
    return height * drho / 1000 * g / 100.0


def _sw_to_simpleJ(sw: float, a: float, b: float) -> float:
    """Convert a water saturation value to the associated J-value,
    using RMS simple-J"""
    return float(a) * sw ** float(b)


def _simpleJ_to_sw(J: float, a: float, b: float) -> float:
    """Convert a J-function-value to a water saturation value,
    using RMS simple-J"""
    return math.pow(J / float(a), 1.0 / float(b))


def _simpleJ_to_height(J: float, poro_ref: float, perm_ref: float) -> float:
    """Convert a J-function value to a height-value in meters

    This scales the J-value with the inverse of characteristic
    length of the capillaries radii, estimated as sqrt(k/phi).

    https://en.wikipedia.org/wiki/Leverett_J-function

    Args:
        J: J-function value.
        poro_ref: Porosity between 0 and 1
        perm_ref: Permeability in milliDarcy
    """
    return J * math.sqrt(float(poro_ref) / float(perm_ref))


def _height_to_simpleJ(H: float, poro_ref: float, perm_ref: float):
    """Convert a height value (in meters) to a corresponding J-function

    This scales the J-value with the characteristic
    length of the capillaries radii, estimated as sqrt(k/phi).

    https://en.wikipedia.org/wiki/Leverett_J-function

    Args:
        H: Height in meters.
        poro_ref: Porosity between 0 and 1
        perm_ref: Permeability in milliDarcy
    """
    return H * math.sqrt(float(perm_ref) / float(poro_ref))


def swl_from_height_simpleJ(
    swlheight: float, swirr: float, a: float, b: float, poro_ref: float, perm_ref: float
) -> float:
    #  disable=invalid-name
    """Calculate a swl value based on a height parameter.

    The height parameter is typically meters above free water level (FWL)

    Args:
        swlheight: Height above free water level, in meters,
            where we want swl to be.
        swirr: Asymptotic irreducible water saturation. This is
            used to normalized the outputted swl.
        a: a coefficient in RMS simplified J function
        b: b coefficient in RMS simplified J function
        poro_ref: Reference porosity for scaling to Pc, between 0 and 1
        perm_ref: Reference permeability for scaling to Pc, in milliDarcy

    Returns:
        A value that can be used for swl.
    """
    # Calculate the value of J() far above free water level:
    j_value_at_swlheight = _height_to_simpleJ(swlheight, poro_ref, perm_ref)

    # Calculate the water saturation value that corresponds to this J-value,
    # un-normalized:
    swn_swlheight = _simpleJ_to_sw(j_value_at_swlheight, a, b)

    # Normalize with respect the the asymptotic swirr and return
    return swirr + (1 - swirr) * swn_swlheight



"""Test module for relperm processing support code in pyscal"""

import numpy as np
import pandas as pd
import pytest

from pyscal.utils.relperm import crosspoint, estimate_diffjumppoint, truncate_zeroness

# pyscal.utils.relperm.crosspoint() is also tested in test_wateroil and test_gasoil.


@pytest.mark.parametrize(
    "value, zeronesslimit, expected",
    [(0, 0.1, 0), (0.01, 0.1, 0), (0.1, 0.1, 0.1), (1, 0.1, 1)],
)
def test_truncate_zeroness(value, zeronesslimit, expected):
    """Test truncation of zeroness, that is that numbers close to zero
    end up as zero, but others don't"""
    assert truncate_zeroness(value, zeronesslimit=zeronesslimit) == expected


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param([[]], -1, id="empty_data"),
        pytest.param(
            [[0, 0, 0]],
            -1,
            id="too_few_rows",
        ),
        pytest.param(
            [[0, 0, 0], [0, 0, 0]],
            0.0,
            id="undefined_crosspoint",
            # Solution is not well defined, but zero is an ok answer.
        ),
        pytest.param(
            [[0, 0, 0], [1, 1, 1]],
            0,  # Should this be detected and -1 returned?
            id="identical_lines",
        ),
        pytest.param(
            [[1, 1, 1], [0, 0, 0]],
            1,  # Should this be detected and -1 returned?
            id="identical_lines",
        ),
        pytest.param(
            [[0, 0, 1], [1, 1, 2]],
            -1,
            id="parallel_lines",
        ),
        pytest.param(
            [[0, 0, 3], [1, 1, 2]],
            -1,
            id="not_crossing",
        ),
        pytest.param(
            [[0, np.nan, 1], [1, 1, 0]],
            -1,
            id="nan_input",
        ),
        pytest.param([[np.nan, 0, 1], [1, 1, 0]], -1, id="nan_input"),
        ([[0, 0, 1], [1, 1, 0]], 0.5),
    ],
)
def test_crosspoint(data, expected):
    """Test the crosspoint computation code.

    This is also tested for relevant data in test_wateroil.py and test_gasoil.py,
    here the possible error scenarios are more explored."""
    if data and data[0]:
        dframe = pd.DataFrame(columns=["A", "B", "C"], data=data)
    else:
        dframe = pd.DataFrame()
    assert np.isclose(
        crosspoint(dframe, "A", "B", "C"),
        expected,
    )


def test_diffjumppoint():
    """Test estimator for the jump in first derivative for some manually set up cases.

    This code is also extensively tested through test_fromtable"""

    dframe = pd.DataFrame(columns=["x", "y"], data=[[0, 0], [0.3, 0.2], [1, 1]])

    assert estimate_diffjumppoint(dframe, side="right") == 0.3
    assert estimate_diffjumppoint(dframe, side="left") == 0.3

    with pytest.raises(AssertionError):
        estimate_diffjumppoint(dframe, side="front")
    with pytest.raises(ValueError):
        estimate_diffjumppoint(dframe, side=None)

    dframe = pd.DataFrame(columns=["x", "y"], data=[[0, 0], [1, 1]])
    # We don't really care what gets printed from this, just don't crash..
    assert 0 <= estimate_diffjumppoint(dframe, side="right") <= 1
    assert 0 <= estimate_diffjumppoint(dframe, side="left") <= 1

    dframe = pd.DataFrame(
        columns=["x", "y"],
        data=[
            [0, 0],
            [0.1, 0.1],
            [0.2, 0.2],  # Linear until here
            [0.3, 0.4],  # Nonlinear region
            [0.4, 0.45],  # Nonlinear region
            [0.7, 0.7],  # Linear from here
            [0.8, 0.8],
            [1, 1],
        ],
    )
    assert estimate_diffjumppoint(dframe, side="left") == 0.2
    assert estimate_diffjumppoint(dframe, side="right") == 0.7

    dframe = pd.DataFrame(
        columns=["x", "y"],
        data=[
            [0, 0],
            [0.1, 0.0],
            [0.2, 0.0],  # Linear until here
            [0.3, 0.4],  # Nonlinear region
            [0.9, 1],  # Start linear region again
            [1, 1],
        ],
    )
    assert estimate_diffjumppoint(dframe, side="left") == 0.2
    assert estimate_diffjumppoint(dframe, side="right") == 0.9



"""Test module endpoints computed by WaterOil"""

import hypothesis.strategies as st
import numpy as np
from hypothesis import given

from pyscal import GasOil, WaterOil
from pyscal.constants import EPSILON as epsilon


def test_swcr():
    """Test that we can locate swcr with a certain accuracy"""
    wateroil = WaterOil(swcr=0.3, swl=0.1, sorw=0.1, h=0.05)
    wateroil.add_corey_oil(now=2, kroend=0.8)
    wateroil.add_corey_water(nw=2, krwend=0.6)
    est_sorw = wateroil.estimate_sorw()
    est_swcr = wateroil.estimate_swcr()

    assert np.isclose(est_sorw, 0.1)
    assert np.isclose(est_swcr, 0.3)


def test_swcrsorw():
    """Test estimate_sorw/swcr for some manually set up cases"""
    # swcr, sorw, h:
    testtuples = [
        (0, 0.3, 0.1),
        (0.01, 0.2, 0.1),
        (0.3, 0.1, 0.1),
        (0.0, 0.0, 0.1),
        (0.00001, 0.9, 0.1),
        (0.0001, 0.99, 0.1),
    ]

    for testtuple in testtuples:
        real_swcr = testtuple[0]
        real_sorw = testtuple[1]
        h = testtuple[2]
        wateroil = WaterOil(swcr=real_swcr, sorw=real_sorw, h=h)
        wateroil.add_corey_oil(now=2, kroend=0.8)
        wateroil.add_corey_water(nw=2, krwend=0.9)
        est_sorw = wateroil.estimate_sorw()
        mis = abs(est_sorw - real_sorw)
        print(f"Testing sorw={real_sorw}, swcr={real_swcr} on h={h}")
        if mis > 0.01:
            print(f"Missed, estimated was {est_sorw}")
        assert mis < h + epsilon  # Can't guarantee better than h.
        est_swcr = wateroil.estimate_swcr()
        mis_swcr = abs(est_swcr - real_swcr)
        if mis_swcr > 0.0:
            print(f"Missed swcr, estimate was {est_swcr}")
        assert mis_swcr < h + epsilon


def test_sorg():
    """Test estimate_sorg for some manually set up cases"""
    # sorg, sgcr, h, swl:
    testtuples = [
        (0.3, 0.01, 0.1, 0.1),
        (0.2, 0, 0.05, 0.0),
        (0.1, 0.3, 0.01, 0.5),
        (0.0, 0, 0.1, 0.000001),
        (0.9, 0.000001, 0.1, 0),
        (0.4, 0, 0.1, 0.2),
    ]

    for testtuple in testtuples:
        real_sorg = testtuple[0]
        real_sgcr = testtuple[1]
        h = testtuple[2]
        swl = testtuple[3]
        gasoil = GasOil(sgcr=0.03, sorg=real_sorg, h=h, swl=swl)
        gasoil.add_corey_oil(nog=2)
        gasoil.add_corey_gas(ng=2, krgend=0.9)
        print(f"Testing sorg={real_sorg} on h={h}, swl={swl}")
        est_sorg = gasoil.estimate_sorg()
        mis = abs(est_sorg - real_sorg)
        if mis > 0.01:
            print(f"Missed, estimated was {est_sorg}")
        assert mis < h + epsilon  # Can't guarantee better than h.

        # If krgendanchor is not sorg (default), then krg cannot be used
        # and the GasOil object will resort to using krog. Should work
        # when now=2 but might not always work for all kinds of LET parameters.
        gasoil = GasOil(sorg=real_sorg, h=h, swl=swl, krgendanchor="")
        gasoil.add_corey_oil(nog=2)
        gasoil.add_corey_gas(ng=2, krgend=0.8)
        est_sorg = gasoil.estimate_sorg()
        print(f"Estimated to {est_sorg}")
        mis = abs(est_sorg - real_sorg)
        assert mis < h + epsilon

        # Test sgcr:
        gasoil = GasOil(
            sorg=real_sorg, sgcr=real_sgcr, h=h, swl=swl, krgendanchor="sorg"
        )
        gasoil.add_corey_oil(nog=2, kroend=0.8)
        gasoil.add_corey_gas(ng=2, krgend=0.8)
        est_sgcr = gasoil.estimate_sgcr()
        mis = abs(est_sgcr - real_sgcr)
        assert mis < h + epsilon


def test_linear_corey():
    """Test how the estimate_sorw works when Corey is linear"""
    # Piecewise linear, this we can detect:
    wateroil = WaterOil(swl=0, h=0.01, sorw=0.3)
    wateroil.add_corey_water(nw=1, krwend=0.8)
    est_sorw = wateroil.estimate_sorw()
    assert np.isclose(est_sorw, 0.3)

    # Piecewise linear, this we can detect:
    wateroil = WaterOil(swl=0, h=0.01, sorw=0.5)
    wateroil.add_corey_water(nw=1, krwend=1)
    est_sorw = wateroil.estimate_sorw()
    assert np.isclose(est_sorw, 0.5)

    # This one is impossible to deduct sorw from, because
    # the entire krw curve is linear.
    wateroil = WaterOil(swl=0, h=0.01, sorw=0.5)
    wateroil.add_corey_water(nw=1, krwend=0.5)
    wateroil.add_corey_oil(now=2)
    est_sorw = wateroil.estimate_sorw()
    # The code will return sw = swl + h by design:
    # Unless information is in the krow table, anything better is impossible
    assert np.isclose(est_sorw, 1 - 0.01)


@given(
    st.floats(min_value=0.001, max_value=0.2),
    st.floats(min_value=0.0001, max_value=0.5),
    st.floats(min_value=0.5, max_value=3),
)
def test_sorw_hypo(h, sorw, nw):
    """Test estimate_sorw extensively"""
    wateroil = WaterOil(sorw=sorw, h=h)
    wateroil.add_corey_oil(now=2)
    wateroil.add_corey_water(nw=nw)
    est_sorw = wateroil.estimate_sorw()
    est_error = abs(sorw - est_sorw)
    assert not np.isnan(est_sorw)
    error_requirement = h + 10000 * epsilon

    if abs(1 - nw) < 0.1:
        # When the Corey curve is almost linear, we don't
        # expect sorw estimation to be robust
        error_requirement = 3 * h
    if est_error > error_requirement:
        print(f"h = {h}, sorw = {sorw}")
        print(f"estimated sorw = {est_sorw}")
        print(wateroil.table)

    # We don't bother to tune the code better than this:
    assert est_error <= error_requirement



"""Test module for capillary pressure in WaterOil"""

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given

from pyscal import WaterOil
from pyscal.constants import MAX_EXPONENT_PC
from pyscal.utils.testing import check_table, float_df_checker, sat_table_str_ok


def test_simple_j():
    """Simple test of the simple J function correlation"""
    wateroil = WaterOil(swl=0.01)
    wateroil.add_simple_J()  # swl set to zero will give infinite pc
    check_table(wateroil.table)
    assert wateroil.pccomment

    wateroil = WaterOil(swl=0)
    with pytest.raises(ValueError, match="swl must be larger than zero"):
        wateroil.add_simple_J()

    # Zero gravity:
    wateroil = WaterOil(swl=0.01)
    wateroil.add_simple_J(g=0)
    assert wateroil.table["PC"].unique() == 0.0

    # This should give back Sw:
    # This ensures that density and gravity scaling is correct
    wateroil.add_simple_J(a=1, b=1, poro_ref=1, perm_ref=1, drho=1000, g=100)
    assert (wateroil.table["PC"] - wateroil.table["SW"]).sum() < 0.00001
    # (check_table() will fail on this, when b > 0)

    # Some values seen in real life:
    wateroil.add_simple_J(a=100, b=-1.5, poro_ref=0.12, perm_ref=100, drho=200)
    check_table(wateroil.table)
    assert "Simplified" in wateroil.pccomment
    assert "a=100" in wateroil.pccomment
    assert "b=-1.5" in wateroil.pccomment
    wateroil.add_corey_oil()
    wateroil.add_corey_water()
    swof = wateroil.SWOF()
    assert isinstance(swof, str)
    assert swof
    sat_table_str_ok(swof)
    sat_table_str_ok(wateroil.SWFN())


def test_simple_j_petro():
    """Simple test of the simple J petrophysical function correlation"""
    wateroil = WaterOil(swl=0)
    with pytest.raises(ValueError, match="swl must be larger than zero"):
        wateroil.add_simple_J_petro(a=1, b=-2)

    wateroil = WaterOil(swl=0.01)
    wateroil.add_simple_J_petro(a=1, b=-2)
    check_table(wateroil.table)
    assert wateroil.pccomment
    assert "etrophysic" in wateroil.pccomment

    # Zero gravity:
    wateroil.add_simple_J_petro(a=1, b=-2, g=0)
    assert wateroil.table["PC"].unique() == 0.0

    # Numerical test from sample numbers calculated independently in different tool:
    wateroil = WaterOil(swl=0.05, h=0.025)
    wateroil.add_simple_J_petro(
        a=1.45, b=-0.285, drho=143, g=9.81, perm_ref=15, poro_ref=0.27
    )
    float_df_checker(wateroil.table, "SW", 0.1, "PC", 22.36746)
    assert "Simplified" in wateroil.pccomment
    assert "etrophysic" in wateroil.pccomment
    wateroil.add_corey_oil()
    wateroil.add_corey_water()
    swof = wateroil.SWOF()
    assert isinstance(swof, str)
    assert swof
    sat_table_str_ok(swof)
    sat_table_str_ok(wateroil.SWFN())

    with pytest.raises(ValueError, match="positive b"):
        wateroil = WaterOil(swl=0.01)
        wateroil.add_simple_J_petro(a=1, b=2)


@given(
    st.floats(min_value=0.001, max_value=1000000),
    st.floats(min_value=-0.9 * MAX_EXPONENT_PC, max_value=-0.001),
    st.floats(min_value=0.01, max_value=0.5),
    st.floats(min_value=0.001, max_value=10),
    st.floats(min_value=0.01, max_value=1000000),
    st.floats(min_value=0.001, max_value=10000000),
)
def test_simple_j_random(a, b, poro_ref, perm_ref, drho, g):
    """Test different J-function parameters.

    Parameter ranges tested through hypothesis are limited, as not
    every number is realistic. Way outside the tested intervals, you
    can get AssertionErrors or the capillary pressure may not be
    monotonically decreasing within machine precision.
    """
    wateroil = WaterOil(swl=0.01)
    wateroil.add_simple_J(
        a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
    )
    check_table(wateroil.table)


def test_normalized_j(caplog):
    """Test the normalized J-function correlation for capillary pressure"""
    wateroil = WaterOil(swirr=0.1, h=0.1)
    with pytest.raises(ValueError):
        wateroil.add_normalized_J(a=0.5, b=-0.2, poro=0.2, perm=10, sigma_costau=30)

    wateroil = WaterOil(swirr=0, swl=0.1, h=0.1)
    wateroil.add_normalized_J(a=0.5, b=-0.2, poro=0.2, perm=10, sigma_costau=30)
    check_table(wateroil.table)

    # Sample numerical tests taken from a prior implementation
    # NB: Prior implementation created Pc in atm, we create in bar
    bar_to_atm = 1.0 / 1.01325
    wateroil.add_normalized_J(a=0.22, b=-0.5, perm=100, poro=0.2, sigma_costau=30)
    float_df_checker(wateroil.table, "SW", 0.1, "PC", 2.039969 * bar_to_atm)
    float_df_checker(wateroil.table, "SW", 0.6, "PC", 0.056666 * bar_to_atm)
    float_df_checker(wateroil.table, "SW", 1.0, "PC", 0.02040 * bar_to_atm)

    wateroil = WaterOil(swirr=0.1, swl=0.11, h=0.1)
    wateroil.add_normalized_J(a=0.5, b=-0.001, poro=0.2, perm=10, sigma_costau=30)
    assert "b exponent is very small" in caplog.text

    wateroil = WaterOil(swirr=0.1, swl=0.11, h=0.1)
    wateroil.add_normalized_J(a=0.001, b=-2, poro=0.2, perm=10, sigma_costau=30)
    assert "a parameter is very small" in caplog.text

    wateroil = WaterOil(swirr=0.1, swl=0.11, h=0.1)
    wateroil.add_normalized_J(a=9, b=-2, poro=0.2, perm=10, sigma_costau=30)
    assert "a parameter is very high" in caplog.text


@given(
    st.floats(min_value=0, max_value=0.1),  # swirr
    st.floats(min_value=0.01, max_value=0.1),  # swl - swirr
    st.floats(min_value=0.01, max_value=5),  # a
    st.floats(min_value=-0.9 * MAX_EXPONENT_PC, max_value=-0.01),  # b
    st.floats(min_value=-1, max_value=1.5),  # poro
    st.floats(min_value=0.0001, max_value=1000000000),  # perm
    st.floats(min_value=0, max_value=100000),  # sigma_costau
)
def test_norm_j_pc_random(swirr, swl, a_pc, b_pc, poro, perm, sigma_costau):
    """Test many possibilities of Pc-parameters.

    Outside of the tested range, there are many combination of parameters
    that can give infinite capillary pressure"""

    swl = swirr + swl  # No point in getting too many AssertionErrors
    wateroil = WaterOil(swirr=swirr, swl=swl, h=0.01)
    try:
        wateroil.add_normalized_J(
            a=a_pc, b=b_pc, perm=perm, poro=poro, sigma_costau=sigma_costau
        )
    except (AssertionError, ValueError):  # when poro is < 0 f.ex.
        return
    check_table(wateroil.table)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    sat_table_str_ok(wateroil.SWOF())


def test_let_pc_pd():
    """Test LET formulation for primary drainage capillary pressure"""
    wateroil = WaterOil(swirr=0.1)
    wateroil.add_LET_pc_pd(Lp=1, Ep=1, Tp=1, Lt=1, Et=1, Tt=1, Pcmax=10, Pct=5)
    assert np.isclose(wateroil.table["PC"].max(), 10)
    assert np.isclose(wateroil.table["PC"].min(), 0)
    # (everything is linear)

    wateroil.add_LET_pc_pd(Lp=10, Ep=10, Tp=10, Lt=10, Et=10, Tt=10, Pcmax=10, Pct=5)
    assert np.isclose(wateroil.table["PC"].max(), 10)
    assert np.isclose(wateroil.table["PC"].min(), 0)
    # On a plot, you can see a kink at Pc=5.
    # wateroil.plotpc()

    wateroil = WaterOil(swirr=0.1, sorw=0.4)
    wateroil.add_LET_pc_pd(Lp=10, Ep=10, Tp=10, Lt=10, Et=10, Tt=10, Pcmax=5, Pct=2)
    assert np.isclose(wateroil.table["PC"].max(), 5)
    assert np.isclose(wateroil.table["PC"].min(), 0)
    # On plot: hard-to-see kink at Pc=2. .
    # wateroil.plotpc()
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    sat_table_str_ok(wateroil.SWOF())


def test_let_pc_imb():
    """Test the LET formulation for imbibition capillary pressures"""
    wateroil = WaterOil(swirr=0.1)
    wateroil.add_LET_pc_imb(
        Ls=1, Es=1, Ts=1, Lf=1, Ef=1, Tf=1, Pcmax=10, Pcmin=-10, Pct=3
    )
    assert np.isclose(wateroil.table["PC"].max(), 10)
    assert np.isclose(wateroil.table["PC"].min(), -10)

    wateroil = WaterOil(swirr=0.1)
    wateroil.add_LET_pc_imb(Ls=5, Es=5, Ts=5, Lf=5, Ef=5, Tf=5, Pcmax=5, Pcmin=1, Pct=4)
    assert np.isclose(wateroil.table["PC"].max(), 5)
    assert np.isclose(wateroil.table["PC"].min(), 1)

    wateroil = WaterOil(swirr=0.1, sorw=0.3)
    wateroil.add_LET_pc_imb(Ls=5, Es=5, Ts=5, Lf=5, Ef=5, Tf=5, Pcmax=5, Pcmin=1, Pct=4)
    assert np.isclose(wateroil.table["PC"].max(), 5)
    assert np.isclose(wateroil.table["PC"].min(), 1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    sat_table_str_ok(wateroil.SWOF())


def test_skjaeveland_pc():
    """Simple test of Skjæveland capillary pressure correlation"""
    wateroil = WaterOil(h=0.3, swl=0.2, swirr=0.1)
    wateroil.add_corey_water()
    wateroil.add_skjaeveland_pc(swr=0.1, cw=0.1, co=-0.1, aw=0.1, ao=0.1)
    check_table(wateroil.table)
    swfn = wateroil.SWFN()

    # Defaulting swr should give the same since swr==swirr above:
    wateroil.add_skjaeveland_pc(swr=None, cw=0.1, co=-0.1, aw=0.1, ao=0.1)
    assert wateroil.SWFN() == swfn

    # Add with wrong numbers
    wateroil = WaterOil(h=0.3, swl=0.2, sorw=0.3)
    with pytest.raises(ValueError, match="cw must be larger"):
        wateroil.add_skjaeveland_pc(swr=0.1, cw=-0.1, co=-0.1, aw=0.1, ao=0.1)

    with pytest.raises(ValueError, match="co must be less"):
        wateroil.add_skjaeveland_pc(swr=0.1, cw=0.1, co=0.1, aw=0.1, ao=0.1)

    with pytest.raises(ValueError, match="aw must be larger"):
        wateroil.add_skjaeveland_pc(swr=0.1, cw=0.1, co=-0.1, aw=-0.1, ao=0.1)

    with pytest.raises(ValueError, match="ao must be larger"):
        wateroil.add_skjaeveland_pc(swr=0.1, sor=0.3, cw=0.1, co=-0.1, aw=0.1, ao=-0.1)

    with pytest.raises(ValueError, match=r"swr \(swirr\) must be less than 1 - sor"):
        wateroil.add_skjaeveland_pc(swr=0.8, sor=0.3, cw=0.1, co=-0.1, aw=0.1, ao=0.1)

    with pytest.raises(ValueError, match="swr must be contained in"):
        wateroil.add_skjaeveland_pc(swr=-1.2, sor=0.3, cw=0.1, co=-0.1, aw=0.1, ao=0.1)

    with pytest.raises(ValueError, match="swr must be contained in"):
        wateroil.add_skjaeveland_pc(swr=1.2, sor=0.3, cw=0.1, co=-0.1, aw=0.1, ao=0.1)

    with pytest.raises(ValueError, match="sor must be contained in"):
        wateroil.add_skjaeveland_pc(swr=0.1, sor=-0.2, cw=0.1, co=-0.1, aw=0.1, ao=0.1)

    with pytest.raises(ValueError, match="sor must be contained in"):
        wateroil.add_skjaeveland_pc(swr=0.1, sor=1.2, cw=0.1, co=-0.1, aw=0.1, ao=0.1)



"""Test module for capillary pressure support code in pyscal"""

import math

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given

from pyscal.constants import EPSILON
from pyscal.utils import capillarypressure

PASCAL = 1e-05  # One pascal in bar.


@pytest.mark.parametrize(
    "sw, a, b, poro_ref, perm_ref, drho, g, expected",
    [
        (0, 0, 0, 0, 1, 0, 0, 0),
        (1, 1, 1, 1, 1, 1, 1, PASCAL),
        # Linear in gravity:
        (1, 1, 1, 1, 1, 1, 10, 10 * PASCAL),
        # Linear in density difference
        (1, 1, 1, 1, 1, 10, 1, 10 * PASCAL),
        # Nonlinear in permeabilty
        (1, 1, 1, 1, 10, 1, 1, math.sqrt(1.0 / 10.0) * PASCAL),
        # Nonlinear in porosity:
        (1, 1, 1, 0.5, 1, 1, 1, math.sqrt(0.5) * PASCAL),
        # Linear in a:
        (1, 2, 1, 1, 1, 1, 1, 2 * PASCAL),
        # Nonlinear in b:
        (1, 1, 2, 1, 1, 1, 1, PASCAL),
        (0.5, 1, 2, 1, 1, 1, 1, 0.5**2 * PASCAL),
        # Vector support in sw:
        (np.array([0, 1]), 1, 1, 1, 1, 1, 1, np.array([0, PASCAL])),
    ],
)
def test_simple_J(sw, a, b, poro_ref, perm_ref, drho, g, expected):
    """Test the simple J formula implementation"""
    result = capillarypressure.simple_J(sw, a, b, poro_ref, perm_ref, drho, g)

    if isinstance(result, (list, np.ndarray)):
        assert np.isclose(result, expected).all()
    else:
        assert np.isclose(result, expected)


@pytest.mark.parametrize(
    "swlheight, swirr, a, b, poro_ref, perm_ref, expected",
    [
        (1, 0, 1, -1, 1, 1, 1),
        (1, 0.1, 1, -1, 1, 1, 1),
    ],
)
def test_swl_from_height_simple_J(swlheight, swirr, a, b, poro_ref, perm_ref, expected):
    """Test the calculation of swlheight from input parameters"""
    result = capillarypressure.swl_from_height_simpleJ(
        swlheight, swirr, a, b, poro_ref, perm_ref
    )

    if isinstance(result, (list, np.ndarray)):
        assert np.isclose(result, expected).all()
    else:
        assert np.isclose(result, expected)


@given(
    st.floats(min_value=EPSILON, max_value=100),
    st.floats(min_value=0.001, max_value=1000000),
    st.floats(min_value=-9, max_value=-0.1),
    # Higher b gives OverflowError: math range error
)
def test_inverses_sw_simpleJ(j_value, a, b):
    """Ensure that the pair of functions going from sw to J and back
    are truly inverses of each other"""
    sw = capillarypressure._simpleJ_to_sw(j_value, a, b)
    assert np.isclose(capillarypressure._sw_to_simpleJ(sw, a, b), j_value)


@given(
    st.floats(min_value=EPSILON, max_value=1),
    st.floats(min_value=0.001, max_value=1000000),
    st.floats(min_value=-9, max_value=-0.1),
)
def test_inverses_simpleJ_sw(sw_value, a, b):
    """Inverse of the test function above"""
    result = capillarypressure._simpleJ_to_sw(
        capillarypressure._sw_to_simpleJ(sw_value, a, b), a, b
    )
    assert np.isclose(result, sw_value)


@given(
    st.floats(min_value=-10000000, max_value=10000000),  # J
    st.floats(min_value=0.0001, max_value=1),  # poro_ref
    st.floats(min_value=0.1, max_value=10000),  # perm_ref
)
def test_inverses_simpleJ_height(J, poro_ref, perm_ref):
    """Test round-trip calculation of J-value from height"""
    result = capillarypressure._height_to_simpleJ(
        capillarypressure._simpleJ_to_height(J, poro_ref, perm_ref), poro_ref, perm_ref
    )
    assert np.isclose(result, J)


@given(
    st.floats(min_value=-100, max_value=100),  # height
    st.floats(min_value=0.0001, max_value=1),  # poro_ref
    st.floats(min_value=0.1, max_value=10000),  # perm_ref
)
def test_inverses_height_simpleJ(height, poro_ref, perm_ref):
    """Test round-trip calculation of height-value from J"""
    result = capillarypressure._simpleJ_to_height(
        capillarypressure._height_to_simpleJ(height, poro_ref, perm_ref),
        poro_ref,
        perm_ref,
    )
    assert np.isclose(result, height)


def test_reference_implementation_swl_from_height():
    """Test the reference implementation that was copied from Drogon example project"""
    swlheight = 300
    permref = 10
    pororef = 0.3
    a = 1
    b = -2.2
    swirr = 0.02

    # Lines copied from Drogon:
    j_swlheight = swlheight * math.sqrt(permref / pororef)  # J at swlheight
    swn_swlheight = math.pow(j_swlheight / a, 1 / b)  # swn at swlheigh_
    ref_swl = swirr + (1 - swirr) * swn_swlheight  # swl = sw at swlheight

    swl = capillarypressure.swl_from_height_simpleJ(
        swlheight, swirr, a, b, pororef, permref
    )

    assert np.isclose(swl, ref_swl)



"""Wateroil module"""

import math
from typing import Optional

import numpy as np
import pandas as pd
from scipy.interpolate import PchipInterpolator, interp1d

import pyscal
from pyscal.constants import EPSILON as epsilon
from pyscal.constants import MAX_EXPONENT_KR, MAX_EXPONENT_PC, SWINTEGERS
from pyscal.utils.capillarypressure import simple_J
from pyscal.utils.relperm import crosspoint, estimate_diffjumppoint, truncate_zeroness
from pyscal.utils.string import comment_formatter, df2str

logger = pyscal.getLogger_pyscal(__name__)


class WaterOil:
    """A representation of two-phase properties for oil-water.

    Can hold relative permeability data, and capillary pressure.

    Parametrizations for relative permeability
     * Corey
     * LET

    For capillary pressure
     * Simplified J-function

    For object initialization, only saturation endpoints must be inputted,
    and saturation resolution. An optional string can be added as a 'tag'
    that can be used when outputting.

    Relative permeability and/or capillary pressure can be added through
    parametrizations, or from a dataframe (will incur interpolation).

    Can be dumped as include files for Eclipse/OPM and Nexus simulators.

    Args:
        swirr: Absolute minimal water saturation at infinite capillary
            pressure.
        swl: First water saturation point in generated table. Used
            for normalized saturations.
        swcr: Critical water saturation. Water will not be mobile before the
            water saturation is above this value.
        sorw: Residual oil saturation after water flooding. At this oil
            saturation, the oil has zero relative permeability.
        socr: Critical oil saturation, oil will not be mobile before the
            oil saturation is above socr. This parameter will default to sorw
            and is to be used larger than sorw for oil paleo zone modelling
            cases.
        h: Saturation step-length in the outputted table.
        tag: Optional string identifier, only used in comments.
        fast: Set to True if in order to skip some integrity checks
            and nice-to-have features. Not needed to set for normal pyscal
            runs, as speed is seldom crucial. Default False
    """

    def __init__(
        self,
        swirr: float = 0.0,
        swl: float = 0.0,
        swcr: float = 0.0,
        sorw: float = 0.0,
        socr: Optional[float] = None,
        h: Optional[float] = None,
        tag: str = "",
        fast: bool = False,
        _sgcr: Optional[float] = None,
        _sgl: Optional[float] = None,
    ) -> None:
        """Sets up the saturation range. Swirr is only relevant
        for the capillary pressure, not for relperm data.

        _sgcr and _sgl are only to be used by the GasWater object.
        """

        assert -epsilon - 1 < swirr < 1.0 + epsilon, (
            f"swirr = {swirr}, -1 <= swirr <= 1 is required"
        )
        assert -epsilon < swl < 1.0 + epsilon, f"swl = {swl}, 0 <= swl < 1 is required"
        assert -epsilon < swcr < 1.0 + epsilon, (
            f"swcr = {swcr}, 0 <= swcr < 1 is required"
        )
        assert -epsilon < sorw < 1.0 + epsilon, (
            f"sorw = {sorw}, 0 <= sorw < 1 is required"
        )
        if socr is not None:
            assert -epsilon < socr < 1.0 + epsilon, "0 <= socr < 1 is required"

        if h is None:
            h = 0.01

        assert swl < 1 - sorw
        assert swcr < 1 - sorw
        assert swirr < 1 - sorw

        if swirr < 0:
            logger.warning(
                "Negative swirr value, %f, detected. Negative values are allowed,"
                " but you should ensure that this is intentional.",
                swirr,
            )

        self.swcomment: str = ""

        h_min = 10.0 / float(SWINTEGERS)
        if h < h_min:
            logger.warning(
                "Requested saturation step length (%g) too small, reset to %g", h, h_min
            )
            self.h = h_min
        else:
            self.h = h

        if _sgcr is not None:
            self.sgcr = _sgcr
        if _sgl is not None:
            assert swl < 1 - _sgl < 1 + epsilon, "1-sgl must be between swl and 1"
            assert self.sgcr + epsilon > _sgl, "sgcr must be larger than sgl"
            assert sorw + epsilon > _sgl, "sorw/sgrw must be larger than sgl"
            self.sgl = truncate_zeroness(_sgl, name="_sgl")
        else:
            self.sgl = 0.0

        self.swirr = swirr
        self.swl = max(swl, swirr)  # Cannot allow swl < swirr. Warn?
        self.sorw = truncate_zeroness(sorw, name="sorw")
        if self.swl < swcr < self.swl + 1 / SWINTEGERS + epsilon:
            # Give up handling swcr so close to swl
            swcr = self.swl
        self.swcr = max(self.swl, swcr)  # Cannot allow swcr < swl. Warn?

        if socr is not None:
            self.socr = truncate_zeroness(socr, name="socr")
            if socr < self.sorw - epsilon:
                raise ValueError(
                    "socr must be equal to or larger than sorw, "
                    f"socr={socr}, sorw={self.sorw}"
                )
            if self.sorw - epsilon < self.socr < self.sorw + 1 / SWINTEGERS + epsilon:
                if self.sorw < self.socr or self.sorw > self.socr:
                    # Only warn if it seems the user actually tried to set socr
                    logger.warning("socr was close to sorw, reset to sorw")
                self.socr = self.sorw
        else:
            self.socr = self.sorw

        self.tag = tag
        self.fast = fast
        sw_list = [
            *list(np.arange(self.swl, 1 - self.sgl, self.h)),
            self.swcr,
            1 - self.sorw,
            1 - self.socr,
            1 - self.sgl,
        ]
        sw_list.sort()  # Using default timsort on nearly sorted data.
        self.table = pd.DataFrame(sw_list, columns=["SW"])

        # Ensure that we do not have sw values that are too close
        # to each other, determined rougly by the distance 1/10000
        self.table["swint"] = list(map(round, self.table["SW"] * SWINTEGERS))
        self.table = self.table.drop_duplicates("swint")

        # Now, sw=1-sorw might be accidentaly dropped, so make sure we
        # have it by replacing the closest value by 1-sorw exactly
        sorwindex = (self.table["SW"] - (1 - self.sorw)).abs().sort_values().index[0]
        self.table.loc[sorwindex, "SW"] = 1 - self.sorw

        # Same for sw=1-socr
        socrindex = (self.table["SW"] - (1 - self.socr)).abs().sort_values().index[0]
        self.table.loc[socrindex, "SW"] = 1 - self.socr

        # Same for sw=swcr:
        swcrindex = (self.table["SW"] - (self.swcr)).abs().sort_values().index[0]
        self.table.loc[swcrindex, "SW"] = self.swcr

        # If sw=1 was dropped, then sorw was close to zero:
        if not np.isclose(self.table["SW"].max(), 1.0 - self.sgl):
            # Add it as an extra row:
            self.table.loc[len(self.table) + 1, "SW"] = 1.0 - self.sgl
            self.table = self.table.sort_values(by="SW")

        self.table = self.table.reset_index()
        self.table = self.table[["SW"]]  # Drop the swint column

        # Normalize for krw:
        self.table["SWN"] = (self.table["SW"] - self.swcr) / (1 - self.swcr - self.sorw)
        # Normalize for krow:
        self.table["SON"] = (1 - self.table["SW"] - self.socr) / (
            1 - self.swl - self.socr
        )

        # Different normalization for Sw used for capillary pressure
        self.table["SWNPC"] = (self.table["SW"] - swirr) / (1 - swirr)

        if _sgcr is None:
            self.swcomment = (
                f"-- swirr={self.swirr:g} swl={self.swl:g} "
                f"swcr={self.swcr:g} sorw={self.sorw:g}"
            )
            if self.socr > sorw:
                self.swcomment += f" socr={self.socr:g}"
        else:
            # When _sgcr is defined, this object is in use by GasWater
            # (sorw is aliased as sgrw, and sgcr is relevant)
            self.swcomment = (
                f"-- swirr={self.swirr:g} swl={self.swl:g} "
                f"swcr={self.swcr:g} sgrw={self.sorw:g} sgcr={self.sgcr:g}"
            )
            if _sgl is not None:
                self.swcomment += f" sgl={self.sgl:g}"
        self.swcomment += "\n"

        self.krwcomment = ""
        self.krowcomment = ""
        self.pccomment = ""

        logger.debug(
            "Initialized WaterOil with %s saturation points", str(len(self.table))
        )

    def add_fromtable(
        self,
        dframe: pd.DataFrame,
        swcolname: str = "SW",
        krwcolname: str = "KRW",
        krowcolname: str = "KROW",
        pccolname: str = "PCOW",
        krwcomment: str = "",
        krowcomment: str = "",
        pccomment: str = "",
        sorw: Optional[float] = None,
        socr: Optional[float] = None,
    ) -> None:
        """Interpolate relpermdata from a dataframe.

        The saturation range with endpoints must be set up beforehand,
        and must be compatible with the tabular input. The tabular
        input will be interpolated to the initialized Sw-table

        If you have krw and krow in different dataframes, call this
        function twice

        Calling function is responsible for checking if any data was
        actually added to the table.

        The relpermdata will be interpolated using a monotone cubic
        interpolator below 1-sorw, and linearly above 1-sorw. Capillary
        pressure data will be interpolated monotone cubicly over the
        entire saturation interval

        The python package ecl2df has a tool for converting Eclipse input
        files to dataframes.

        Args:
            dframe: containing data
            swcolname: column name with the saturation data in the dataframe df
            krwcolname: name of column in df with krw
            krowcolname: name of column in df with krow
            pccolname: name of column in df with capillary pressure data
            krwcomment: Inserted into comment
            krowcomment: Inserted into comment
            pccomment: Inserted into comment
            sorw: Explicit sorw. If None, it will be estimated from
                the numbers in krw
            socr: Explicit socr. If None, it will be estimated from
                the numbers in krow
        """
        # Avoid having to deal with multi-indices:
        if len(dframe.index.names) > 1:
            logger.warning(
                "add_fromtable() did a reset_index(), consider not supplying MultiIndex"
            )
            dframe = dframe.reset_index()

        if swcolname not in dframe:
            logger.critical(
                "%s not found in dataframe, can't read table data", swcolname
            )
            raise ValueError

        # Typecheck/convert all numerical columns:
        for col in [swcolname, krwcolname, krowcolname, pccolname]:
            if col in dframe and not pd.api.types.is_numeric_dtype(dframe[col]):
                # Try to convert to numeric type
                try:
                    dframe[col] = dframe[col].astype(float)
                    logger.info("Converted column %s to numbers for fromtable()", col)
                except (TypeError, ValueError) as err:
                    raise ValueError(
                        f"Failed to parse column {col} as numbers for add_fromtable()"
                    ) from err

        if (dframe[swcolname].diff() < 0).any():
            raise ValueError("SW data not sorted")
        if sorw is None and krwcolname in dframe:
            sorw = float(dframe[swcolname].max()) - estimate_diffjumppoint(
                dframe, xcol=swcolname, ycol=krwcolname, side="right"
            )
            logger.info("Estimated sorw in tabular data to %f", sorw)
        else:
            sorw = 0
        if krwcolname in dframe:
            assert -epsilon <= sorw <= 1 + epsilon
            if dframe[krwcolname].max() > 1.0:
                raise ValueError("KRW is above 1 in incoming table")
            if dframe[krwcolname].min() < 0.0:
                raise ValueError("KRW is below 0 in incoming table")
            linearpart = dframe[swcolname] >= 1 - sorw
            nonlinearpart = dframe[swcolname] <= 1 - sorw  # (overlapping at sorw)
            if sum(linearpart) < 2:
                # A linear section of length 1 is not a linear section,
                # recategorize as nonlinear:
                linearpart = pd.Series([False] * len(linearpart))
                nonlinearpart = ~linearpart
                sorw = 0
            if sum(nonlinearpart) < 2:
                # A nonlinear section of length 1 is not a nonlinear section,
                # recategorize as linear:
                nonlinearpart = pd.Series([False] * len(nonlinearpart))
                linearpart = ~nonlinearpart
                sorw = 1 - float(self.table["SW"].min())
            if not np.isclose(dframe[swcolname].min(), self.table["SW"].min()):
                raise ValueError("Incompatible swl")
            # Verify that incoming data is increasing (or level):
            if not (dframe[krwcolname].diff().dropna() > -epsilon).all():
                raise ValueError("Incoming KRW not increasing")
            if sum(nonlinearpart) >= 2:
                pchip = PchipInterpolator(
                    dframe[nonlinearpart][swcolname].astype(float),
                    dframe[nonlinearpart][krwcolname].astype(float),
                )
                self.table.loc[self.table["SW"] <= 1 - sorw, "KRW"] = pchip(
                    self.table.loc[self.table["SW"] <= 1 - sorw, "SW"]
                )
            if sum(linearpart) >= 2:
                linearinterpolator = interp1d(
                    dframe[linearpart][swcolname].astype(float),
                    dframe[linearpart][krwcolname].astype(float),
                )
                self.table.loc[self.table["SW"] >= 1 - sorw, "KRW"] = (
                    linearinterpolator(
                        self.table.loc[self.table["SW"] >= 1 - sorw, "SW"]
                    )
                )
            self.table["KRW"] = self.table["KRW"].clip(lower=0.0, upper=1.0)
            self.sorw = sorw
            self.krwcomment = "-- krw from tabular input" + krwcomment + "\n"
            self.swcr = self.estimate_swcr()

        if krowcolname in dframe:
            if socr is None:
                socr = float(dframe[swcolname].max()) - estimate_diffjumppoint(
                    dframe, xcol=swcolname, ycol=krowcolname, side="right"
                )
                if socr > sorw + epsilon:
                    logger.info("Estimated socr in tabular data from krow to %s", socr)
                else:
                    socr = sorw
            assert -epsilon <= socr <= 1 + epsilon
            linearpart = dframe[swcolname] >= 1 - socr
            nonlinearpart = dframe[swcolname] <= 1 - socr  # (overlapping at sorw)
            if sum(linearpart) < 2:
                linearpart = pd.Series([False] * len(linearpart))
                nonlinearpart = ~linearpart
                socr = 0
            if sum(nonlinearpart) < 2:
                nonlinearpart = pd.Series([False] * len(nonlinearpart))
                linearpart = ~nonlinearpart
                socr = 1 - float(self.table["SW"].min())
            if not np.isclose(dframe[swcolname].min(), self.table["SW"].min()):
                raise ValueError("Incompatible swl")
            if not (dframe[krowcolname].diff().dropna() < epsilon).all():
                raise ValueError("Incoming KROW Not decreasing")
            if dframe[krowcolname].max() > 1.0:
                raise ValueError("KROW is above 1 in incoming table")
            if dframe[krowcolname].min() < 0.0:
                raise ValueError("KROW is below 0 in incoming table")
            if sum(nonlinearpart) >= 2:
                pchip = PchipInterpolator(
                    dframe.loc[nonlinearpart, swcolname].astype(float),
                    dframe.loc[nonlinearpart, krowcolname].astype(float),
                )
                self.table.loc[self.table["SW"] <= 1 - socr, "KROW"] = pchip(
                    self.table.loc[self.table["SW"] <= 1 - socr, "SW"]
                )
            if sum(linearpart) >= 2:
                linearinterpolator = interp1d(
                    dframe.loc[linearpart, swcolname].astype(float),
                    dframe.loc[linearpart, krowcolname].astype(float),
                )
                self.table.loc[self.table["SW"] >= 1 - socr, "KROW"] = (
                    linearinterpolator(
                        self.table.loc[self.table["SW"] >= 1 - socr, "SW"]
                    )
                )
            self.table["KROW"] = self.table["KROW"].clip(lower=0.0, upper=1.0)
            self.socr = socr
            self.krowcomment = "-- krow from tabular input" + krowcomment + "\n"
        if pccolname in dframe:
            # Incoming dataframe must cover the range:
            if dframe[swcolname].min() > self.table["SW"].min():
                raise ValueError("Too large swl for pc interpolation")
            if dframe[swcolname].max() < self.table["SW"].max():
                raise ValueError("max(sw) of incoming data not large enough")
            if np.isinf(dframe[pccolname]).any():
                logger.warning(
                    (
                        "Infinity pc values detected. Will be dropped. "
                        "Risk of extrapolation"
                    )
                )
            dframe = dframe.replace([np.inf, -np.inf], np.nan)
            dframe = dframe.dropna(subset=[pccolname], how="all")
            # If nonzero, then it must be decreasing:
            if (
                dframe[pccolname].abs().sum() > 0
                and not (dframe[pccolname].diff().dropna() < 0.0).all()
            ):
                raise ValueError("Incoming pc not decreasing")
            pchip = PchipInterpolator(
                dframe[swcolname].astype(float), dframe[pccolname].astype(float)
            )
            self.table["PC"] = pchip(self.table["SW"])
            if np.isnan(self.table["PC"]).any() or np.isinf(self.table["PC"]).any():
                raise ValueError("inf/nan in interpolated data, check input")
            self.pccomment = "-- pc from tabular input" + pccomment + "\n"

    def add_corey_water(
        self, nw: float = 2.0, krwend: float = 1.0, krwmax: Optional[float] = None
    ) -> None:
        """Add krw data through the Corey parametrization

        A column named 'krw' will be added. If it exists, it will
        be replaced.

        The Corey model applies for sw < 1 - sorw. For higher
        water saturations, krw is linear between krwend and krwmax.

        krwmax will be ignored if sorw is close to zero

        Args:
            nw: Corey parameter for water.
            krwend: value of krw at 1 - sorw.
            krwmax): maximal value at Sw=1. Default 1
        """
        assert 10 * epsilon < nw < MAX_EXPONENT_KR
        if krwmax:
            assert 0 < krwend <= krwmax <= 1.0
        else:
            assert 0 < krwend <= 1.0

        self.table["KRW"] = krwend * self.table["SWN"] ** nw

        self.set_endpoints_linearpart_krw(krwend, krwmax)

        if not krwmax:
            krwmax = 1
        self.krwcomment = (
            f"-- Corey krw, nw={nw:g}, krwend={krwend:g}, krwmax={krwmax:g}\n"
        )

    def set_endpoints_linearpart_krw(
        self, krwend: float, krwmax: Optional[float] = None
    ) -> None:
        """Set linear parts of krw outside endpoints.

        Curve will be linear from [1 - sorw, 1] (from krwmax to krwend)
        and zero in [swl, swcr].

        This function is used by add_corey_water(), and perhaps by other
        utility functions. It should not be necessary for end-users.

        Args:
            krwend: krw at 1 - sorwr
            krwmax: krw at Sw=1. Default 1.
        """
        # The rows and indices involved in the linear section [1-sorw, 1]:
        linear_section_rows = self.table["SW"] > (1 - self.sorw - epsilon)
        linear_section_indices = self.table[linear_section_rows].index
        # (these lists are never empty)

        # Set krwend always (overrides krwmax if sorw=0)
        self.table.loc[linear_section_indices[0], "KRW"] = krwend

        if len(linear_section_indices) > 1:
            if krwmax is None:
                krwmax = 1
            self.table.loc[linear_section_indices[-1], "KRW"] = krwmax
        elif krwmax is not None:
            logger.info("krwmax ignored when sorw is zero")

        # If the linear section is longer than two rows, do linear
        # interpolation inside for krw:
        if len(linear_section_indices) > 2:
            self.table.loc[linear_section_indices[1:-1], "KRW"] = np.nan
            interp_krw = (
                self.table[["SW", "KRW"]]
                .set_index("SW")
                .interpolate(method="index")["KRW"]
            )
            self.table.loc[:, "KRW"] = interp_krw.to_numpy()

        # Left linear section is all zero:
        self.table.loc[self.table["SW"] < self.swcr, "KRW"] = 0

    def set_endpoints_linearpart_krow(self, kroend: float) -> None:
        """Set linear parts of krow outside endpoints

        Curve will be zero in [1 - socr, 1].

        This function is used by add_corey_water(), and perhaps by other
        utility functions. It should not be necessary for end-users.

        Args:
            kroend: value of kro at swcr
        """
        # Set to zero above socr (usually equal to sorw):
        self.table.loc[self.table["SW"] > 1 - self.socr - epsilon, "KROW"] = 0

        # Floating point issues can cause this to have become
        # slightly bigger than krowend.
        self.table.loc[self.table["KROW"] > kroend, "KROW"] = kroend

    def add_LET_water(
        self,
        l: float = 2.0,
        e: float = 2.0,
        t: float = 2.0,
        krwend: float = 1.0,
        krwmax: Optional[float] = None,
    ) -> None:
        """Add krw data through LET parametrization

        The LET model applies for sw < 1 - sgrw. For higher
        water saturations, krw is linear between krwend and krwmax.

        krwmax will be ignored if sorw is close to zero.

        Args:
            l: LET parameter
            e: LET parameter
            t: LET parameter
            krwend: value of krw at 1 - sorw
            krwmax: maximal value at Sw=1. Default 1
        """
        # Similar code in gasoil.add_LET_gas, but readability is
        # better by having them separate.
        assert epsilon < l < MAX_EXPONENT_KR
        assert epsilon < e < MAX_EXPONENT_KR
        assert epsilon < t < MAX_EXPONENT_KR
        if krwmax:
            assert 0 < krwend <= krwmax <= 1.0
        else:
            assert 0 < krwend <= 1.0

        self.table["KRW"] = (
            krwend
            * self.table["SWN"] ** l
            / ((self.table["SWN"] ** l) + e * (1 - self.table["SWN"]) ** t)
        )
        # This equation is undefined for t a float and swn=1, set explicitly:
        self.table.loc[np.isclose(self.table["SWN"], 1.0), "KRW"] = krwend

        self.set_endpoints_linearpart_krw(krwend, krwmax)

        if not krwmax:
            krwmax = 1
        self.krwcomment = (
            f"-- LET krw, l={l:g}, e={e:g}, t={t:g}, "
            f"krwend={krwend:g}, krwmax={krwmax:g}\n"
        )

    def add_LET_oil(
        self,
        l: float = 2.0,
        e: float = 2.0,
        t: float = 2.0,
        kroend: float = 1,
    ) -> None:
        """
        Add kro data through LET parametrization

        Args:
            l: LET parameter
            e: LET parameter
            t: LET parameter
            kroend: value of kro at swl

        Returns:
            None (modifies object)
        """
        assert epsilon < l < MAX_EXPONENT_KR
        assert epsilon < e < MAX_EXPONENT_KR
        assert epsilon < t < MAX_EXPONENT_KR
        assert 0 < kroend <= 1.0

        self.table["KROW"] = (
            kroend
            * self.table["SON"] ** l
            / ((self.table["SON"] ** l) + e * (1 - self.table["SON"]) ** t)
        )
        # This equation is undefined for t a float and son=1, set explicitly:
        self.table.loc[np.isclose(self.table["SON"], 1.0), "KROW"] = kroend

        self.table.loc[self.table["SW"] >= (1 - self.sorw), "KROW"] = 0

        self.set_endpoints_linearpart_krow(kroend)

        self.krowcomment = (
            f"-- LET krow, l={l:g}, e={e:g}, t={t:g}, kroend={kroend:g}\n"
        )

    def add_corey_oil(self, now: float = 2.0, kroend: float = 1.0) -> None:
        """Add kro data through the Corey parametrization

        Corey applies to the interval between swcr and 1 - sorw

        Curve is linear between swl and swcr, zero above 1 - sorw.

        Args:
            now: Corey exponent
            kroend: kro value at swcr
        Returns:
            None (modifies object)
        """
        assert epsilon < now < MAX_EXPONENT_KR
        assert 0 < kroend <= 1.0

        self.table["KROW"] = kroend * self.table["SON"] ** now
        self.table.loc[self.table["SW"] >= (1 - self.sorw), "KROW"] = 0

        self.set_endpoints_linearpart_krow(kroend)

        self.krowcomment = f"-- Corey krow, now={now:g}, kroend={kroend:g}\n"

    def add_simple_J(
        self,
        a: float = 5.0,
        b: float = -1.5,
        poro_ref: float = 0.25,
        perm_ref: float = 100,
        drho: float = 300,
        g: float = 9.81,
    ) -> None:
        r"""Add capillary pressure function from a simplified J-function

        This is the RMS version of the coefficients *a* and *b*, the formula
        used is

        .. math::

            J = a S_w^b

        *J* is not dimensionless in this equation. The capillary pressure will
        be in bars.

        This is identical to the also seen formula

        .. math::

           J = 10^{b \log(S_w) + \log(a)}

        :math:`S_w` in this formula is normalized with respect to the *swirr* variable
        of the WaterOil object.

        Args:
            a: a coefficient
            b: b coefficient
            poro_ref: Reference porosity for scaling to Pc, between 0 and 1
            perm_ref: Reference permeability for scaling to Pc, in milliDarcy
            drho: Density difference between water and oil, in SI units kg/m³.
                Default value is 300
            g: Gravitational acceleration, in SI units m/s²,
                default value is 9.81

        Returns:
            None. Modifies pc column in self.table, using bar as pressure unit.
        """
        assert g >= 0
        assert b < MAX_EXPONENT_PC
        assert b > -MAX_EXPONENT_PC
        assert 0.0 <= poro_ref <= 1.0
        assert perm_ref > 0.0

        if self.swl < epsilon:
            raise ValueError(
                "swl must be larger than zero to avoid infinite capillary pressure"
            )

        if b > 0:
            logger.warning(
                "positive b will give increasing capillary pressure with saturation"
            )

        # swnpc is a normalized saturation, but normalized with
        # respect to swirr, not to swl (the swirr here is sometimes
        # called 'swirra' - asymptotic swirr)

        self.table["PC"] = simple_J(
            self.table["SWNPC"], a, b, poro_ref, perm_ref, drho, g
        )
        self.pccomment = (
            "-- Simplified J-function for Pc; rms version, in bar\n--   "
            f"a={a:g}, b={b:g}, poro_ref={poro_ref:g}, perm_ref={perm_ref:g} mD,"
            f" drho={drho:g} kg/m^3, g={g:g} m/s^2\n"
        )

    def add_simple_J_petro(
        self,
        a: float,
        b: float,
        poro_ref: float = 0.25,
        perm_ref: float = 100,
        drho: float = 300,
        g: float = 9.81,
    ) -> None:
        r"""Add capillary pressure function from a simplified J-function

        This is the *petrophysical* version of the coefficients *a* and *b*, the formula
        used is

        .. math::

            J = \left(\frac{S_w}{a}\right)^{\frac{1}{b}}


        which is identical to

        .. math::

            J = 10^\frac{\log(S_w) - \log(a)}{b}

        *J* is not dimensionless in this equation.

        :math:`S_w` in this formula is normalized with respect to the *swirr* variable
        of the WaterOil object.

        Args:
            a: a coefficient, petrophysical version
            b: b coefficient, petrophysical version
            poro_ref: Reference porosity for scaling to Pc, between 0 and 1
            perm_ref: Reference permeability for scaling to Pc, in milliDarcy
            drho: Density difference between water and oil, in SI units kg/m³.
                Default value is 300
            g: Gravitational acceleration, in SI units m/s²,
                default value is 9.81

        Returns:
            None. Modifies pc column in self.table, using bar as pressure unit.
        """
        assert g >= 0
        assert b < MAX_EXPONENT_PC
        assert b > -MAX_EXPONENT_PC
        assert 0.0 <= poro_ref <= 1.0
        assert perm_ref > 0.0

        if self.swl < epsilon:
            raise ValueError(
                "swl must be larger than zero to avoid infinite capillary pressure"
            )

        if b > 0:
            raise ValueError(
                "positive b will give increasing capillary pressure with saturation"
            )

        # Convert from "Petrophysical" a's and b's to "RMS" a's and b's:
        rms_a = math.pow(1 / a, 1 / b)
        rms_b = 1 / b

        # Use the other variant of this function for actual computation
        self.add_simple_J(rms_a, rms_b, poro_ref, perm_ref, drho, g)

        self.pccomment = (
            "-- Simplified J-function for Pc, petrophysical version, in bar \n--   "
            f"a={a:g}, b={b:g}, poro_ref={poro_ref:g}, perm_ref={perm_ref:g} mD, "
            f"drho={drho:g} kg/m^3, g={g:g} m/s^2\n"
        )

    def add_normalized_J(
        self, a: float, b: float, poro: float, perm: float, sigma_costau: float
    ) -> None:
        # Don't make this a raw string to avoid the \l warning,
        # it destroys the Latex-formatting in sphinx
        r"""
        Add capillary pressure in bar through a normalized J-function.

        .. math::

            p_c = \frac{\left(\frac{S_w}{a}\right)^{\frac{1}{b}}
            \sigma \cos \tau}{\sqrt{\frac{k}{\phi}}}

        The :math:`S_w` saturation used in the formula is normalized with respect
        to the swirr parameter.

        Args:
            a: a parameter
            b: b exponent (typically negative)
            poro: Porosity value, fraction between 0 and 1
            perm: Permeability value in mD
            sigma_costau: Interfacial tension in mN/m (typical value 30 mN/m)

        Returns:
            None. Modifies pc column in self.table, using bar as pressure unit.
        """
        assert epsilon < abs(a) < MAX_EXPONENT_PC
        assert epsilon < abs(b) < MAX_EXPONENT_PC
        assert epsilon < poro <= 1.0
        assert epsilon < perm
        assert isinstance(sigma_costau, (int, float))

        if b < 0 and np.isclose(self.swirr, self.swl):
            raise ValueError("swl must be larger than swirr to avoid infinite p_c")

        if abs(b) < 0.01:
            logger.warning(
                "b exponent is very small, risk of infinite capillary pressure"
            )

        if abs(a) < 0.01:
            logger.warning(
                "a parameter is very small, risk of infinite capillary pressure"
            )

        if abs(a) > 5:
            logger.warning(
                "a parameter is very high, risk of infinite capillary pressure"
            )

        # 1 atm is equivalent to 101325 pascal = 1.01325 bar
        # pascal_to_atm = 1.0 / 101325.0  # = 9.86923267e-6
        pascal_to_bar = 1e-5

        perm_darcy = perm / 1000
        perm_sq_meters = perm_darcy * 9.869233e-13
        tmp = (self.table["SWNPC"] / a) ** (1.0 / b)
        tmp = tmp / math.sqrt(perm_sq_meters / poro)
        tmp = tmp * sigma_costau / 1000  # Converting mN/m to N/m
        self.table["PC"] = tmp * pascal_to_bar
        self.pccomment = (
            "-- Capillary pressure from normalized J-function, in bar\n"
            f"-- a={a:g}, b={b:g}, poro={poro:g}, perm={perm:g} mD, "
            f"sigma_costau={sigma_costau:g} mN/m\n"
        )

    def add_skjaeveland_pc(
        self,
        cw: float,
        co: float,
        aw: float,
        ao: float,
        swr: Optional[float] = None,
        sor: Optional[float] = None,
    ):
        """Add capillary pressure from the Skjæveland correlation,

        Doc: https://wiki.equinor.com/wiki/index.php/Res:The_Skjaeveland_correlation_for_capillary_pressure

        The implementation is unit independent, units are contained in the
        input constants.

        If swr and sor are not provided, it will be taken from the
        swirr and sorw. Only use different values here if you know
        what you are doing.

        Modifies or adds self.table["PC"] if succesful.
        Returns false if error occured.

        """
        if cw < 0:
            raise ValueError("cw must be larger or equal to zero")
        if co > 0:
            raise ValueError("co must be less than zero")
        if aw <= 0:
            raise ValueError("aw must be larger than zero")
        if ao <= 0:
            raise ValueError("ao must be larger than zero")

        if swr is None:
            swr = self.swirr
        if sor is None:
            sor = self.sorw

        if swr < 0 or swr > 1:
            raise ValueError("swr must be contained in [0,1]")
        if sor < 0 or sor > 1:
            raise ValueError("sor must be contained in [0,1]")
        if swr >= 1 - sor:
            raise ValueError("swr (swirr) must be less than 1 - sor")

        self.pccomment = (
            "-- Skjæveland correlation for Pc;\n"
            f"-- cw={cw:g}, co={co:g}, aw={aw:g}, ao={ao:g}, swr={swr:g}, sor={sor:g}\n"
        )

        # swnpc is a normalized saturation, but normalized with
        # respect to swirr, not to swl (the swirr here is sometimes
        # called 'swirra' - asymptotic swirr)

        # swnpc is generated upon object initialization, but overwritten
        # here to most likely the same values.
        self.table["SWNPC"] = (self.table["SW"] - swr) / (1 - swr)

        # sonpc is almost like 'son', but swl is not used here:
        self.table["SONPC"] = (1 - self.table["SW"] - sor) / (1 - sor)

        # The Skjæveland correlation
        self.table.loc[self.table["SW"] < 1 - sor, "PC"] = cw / (
            self.table["SWNPC"] ** aw
        ) + co / (self.table["SONPC"] ** ao)

        # From 1-sor, the pc is not defined. Extrapolate constantly, and let
        # the non-monotonicity be fixed in the output generators.
        self.table["PC"] = self.table["PC"].ffill()

    def add_LET_pc_pd(
        self,
        Lp: float,
        Ep: float,
        Tp: float,
        Lt: float,
        Et: float,
        Tt: float,
        Pcmax: float,
        Pct: float,
    ) -> None:
        """Add a primary drainage LET capillary pressure curve.

        Docs: https://wiki.equinor.com/wiki/index.php/Res:The_LET_correlation_for_capillary_pressure

        Note that Pc where Sw > 1 - sorw will appear linear because
        there are no saturation points in that interval.
        """
        assert epsilon < Lp < MAX_EXPONENT_PC
        assert epsilon < Ep < MAX_EXPONENT_PC
        assert epsilon < Tp < MAX_EXPONENT_PC
        assert epsilon < Lt < MAX_EXPONENT_PC
        assert epsilon < Et < MAX_EXPONENT_PC
        assert epsilon < Tt < MAX_EXPONENT_PC
        assert Pct <= Pcmax

        # The "forced part"
        self.table["Ffpcow"] = (1 - self.table["SWNPC"]) ** Lp / (
            (1 - self.table["SWNPC"]) ** Lp + Ep * self.table["SWNPC"] ** Tp
        )

        # The gradual rise part:
        self.table["Ftpcow"] = self.table["SWNPC"] ** Lt / (
            self.table["SWNPC"] ** Lt + Et * (1 - self.table["SWNPC"]) ** Tt
        )

        # Putting it together:
        self.table["PC"] = (
            (Pcmax - Pct) * self.table["Ffpcow"] - Pct * self.table["Ftpcow"] + Pct
        )

        # Special handling of the interval [0,swirr]
        self.table.loc[self.table["SWN"] < epsilon, "PC"] = Pcmax
        self.pccomment = (
            "-- LET correlation for primary drainage Pc;\n"
            f"-- Lp={Lp:g}, Ep={Ep:g}, Tp={Tp:g}, "
            f"Lt={Lt:g}, Et={Et:g}, Tt={Tt:g}, "
            f"Pcmax={Pcmax:g}, Pct={Pct:g}\n"
        )

    def add_LET_pc_imb(
        self,
        Ls: float,
        Es: float,
        Ts: float,
        Lf: float,
        Ef: float,
        Tf: float,
        Pcmax: float,
        Pcmin: float,
        Pct: float,
    ) -> None:
        """Add an imbition LET capillary pressure curve.

        Docs: https://wiki.equinor.com/wiki/index.php/Res:The_LET_correlation_for_capillary_pressure
        """
        assert epsilon < Ls < MAX_EXPONENT_PC
        assert epsilon < Es < MAX_EXPONENT_PC
        assert epsilon < Ts < MAX_EXPONENT_PC
        assert epsilon < Lf < MAX_EXPONENT_PC
        assert epsilon < Ef < MAX_EXPONENT_PC
        assert epsilon < Tf < MAX_EXPONENT_PC
        assert Pcmin <= Pct <= Pcmax

        # Normalized water saturation including sorw
        self.table["SWNPCO"] = (self.table["SW"] - self.swirr) / (
            1 - self.sorw - self.swirr
        )

        # The "forced part"
        self.table["Fficow"] = self.table["SWNPCO"] ** Lf / (
            self.table["SWNPCO"] ** Lf + Ef * (1 - self.table["SWNPCO"]) ** Tf
        )

        # The spontaneous part:
        self.table["Fsicow"] = (1 - self.table["SWNPCO"]) ** Ls / (
            (1 - self.table["SWNPCO"]) ** Ls + Es * self.table["SWNPCO"] ** Ts
        )

        # Putting it together:
        self.table["PC"] = (
            (Pcmax - Pct) * self.table["Fsicow"]
            + (Pcmin - Pct) * self.table["Fficow"]
            + Pct
        )

        # Special handling of the interval [0,swirr]
        self.table.loc[self.table["SWNPCO"] < epsilon, "PC"] = Pcmax
        # and [1-sorw,1]
        self.table.loc[self.table["SWNPCO"] > 1 - epsilon, "PC"] = Pcmin
        self.pccomment = (
            "-- LET correlation for imbibition Pc;\n"
            f"-- Ls={Ls:g}, Es={Es:g}, Ts={Ts:g}, "
            f"Lf={Lf:g}, Ef={Ef:g}, Tf={Tf:g}, "
            f"Pcmax={Pcmax:g}, Pcmin={Pcmin:g}, Pct={Pct:g}\n"
        )

    def estimate_sorw(self, curve: str = "KRW") -> float:
        """Estimate sorw of the current krw data.

        This is mostly relevant when add_fromtable() has been used.
        sorw is estimated by searching for a linear part in krw downwards from sw=1.
        In practice it is impossible to infer sorw = 0, since we are limited by
        h, and the last segment from sw=1-h to sw=1 can always be assumed linear.
        Expect sorw = h if the real sorw = 0, but do not depend that it might
        not return zero in the future (one could argue that sorw = h should be
        specially treated to mean sorw = 0)

        If the curve is linear everywhere, sorw will be returned as 1 - (swl + h)

        krow is not used, and should probably not be, as it can be very close to
        zero before approaching sorw.

        Args:
            curve: Colum name of column to use, default is krw.
                If this is all linear, but krow is not, you might be better off
                with krow
        Returns:
            The estimated sorw.
        """
        assert curve in self.table
        assert self.table[curve].sum() > 0
        return self.table["SW"].max() - estimate_diffjumppoint(
            self.table, xcol="SW", ycol=curve, side="right"
        )

    def estimate_socr(self) -> float:
        """Estimate socr from the current kro data."""
        assert self.table["KROW"].sum() > 0
        return self.table["SW"].max() - estimate_diffjumppoint(
            self.table, xcol="SW", ycol="KROW", side="right"
        )

    def estimate_swcr(self, curve: str = "KRW") -> float:
        """Estimate swcr of the current krw data.

        kwcr is estimated by searching for a linear part in krw upwards from sw=swl.
        In practice it is impossible to infer swcr = 0, since we are limited by
        h, and the first segment is assumed linear anyways.

        If the curve is linear everywhere, swcr can end up at the right end
        of your saturation interval.

        Args:
            curve: Colum name of column to use, default is krow.
                If this is all linear, but krw is not, you might be better off
                with krw
        Returns:
            The estimated sgcr.
        """
        assert curve in self.table
        assert self.table[curve].sum() > 0
        return estimate_diffjumppoint(self.table, xcol="SW", ycol=curve, side="left")

    def crosspoint(self) -> float:
        """Locate and return the saturation point where krw = krow

        Accuracy of this crosspoint depends on the resolution chosen
        when initializing the saturation range

        Returns:
            The water saturation where krw == krow, for relperm
            linearly interpolated in water saturation.
        """
        return crosspoint(self.table, "SW", "KRW", "KROW")

    def selfcheck(self, mode: str = "SWOF") -> bool:
        """Check validities of the data in the table.

        An unfinished object will return False.

        If you call SWOF, this function must not return False

        This function should not throw an exception, but capture
        the error and give an error.

        Args:
            mode: "SWOF" or "SWFN". If SWFN, krow is not required.
        """
        error = False
        if "KRW" not in self.table:
            logger.error("krw data not found")
            error = True
        if not (self.table["SW"].diff().dropna().round(10) > -epsilon).all():
            logger.error("SW data not strictly increasing")
            error = True
        if (
            "KRW" in self.table
            and not (self.table["KRW"].diff().dropna().round(10) >= -epsilon).all()
        ):
            logger.error("KRW data not monotonically increasing")
            error = True
        if mode != "SWFN":
            if "KROW" not in self.table:
                logger.error("KROW data not found")
                error = True

            if (
                "KROW" in self.table.columns
                and not (self.table["KROW"].diff().dropna().round(10) <= epsilon).all()
            ):
                # In normal Eclipse runs, krow needs to be level or decreasing.
                # In hysteresis runs, it needs to be strictly decreasing, that must
                # be the users responsibility.
                logger.error("KROW data not level or monotonically decreasing")
                error = True
        if (
            "PC" in self.table.columns
            and self.table["PC"][0] > -epsilon
            and not (self.table["PC"].diff().dropna().round(10) < epsilon).all()
        ):
            logger.error("PC data not strictly decreasing")
            error = True
        if "PC" in self.table.columns and np.isnan(self.table["PC"]).any():
            logger.error("pc data contains NaN")
            error = True
        if "PC" in self.table.columns and np.isinf(self.table["PC"].max()):
            logger.error("pc goes to infinity. Maybe swirr=swl?")
            error = True
        for col in list({"SW", "KRW", "KROW"} & set(self.table.columns)):
            if not (
                (round(min(self.table[col]), 10) >= -epsilon)
                and (round(max(self.table[col]), 10) <= 1 + epsilon)
            ):
                logger.error("%s data should be contained in [0,1]", col)
                error = True
        if error:
            return False
        logger.debug("WaterOil object is checked to be valid")
        return True

    def SWOF(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """
        Produce SWOF input for Eclipse reservoir simulator.

        The columns sw, krw, krow and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: Indicate whether the SWOF string should
                be emitted. If you have multiple SATNUMs, you should
                set this to True only for the first (or False for all,
                and emit the SWOF yourself). Default True
            dataincommentrow: Wheter metadata should
                be printed. Defualt True

        """
        if not self.fast and not self.selfcheck():
            # selfcheck failed and has issued an error message
            return ""
        string = ""
        if header:
            string += "SWOF\n"
        string += comment_formatter(self.tag)
        string += "-- pyscal: " + str(pyscal.__version__) + "\n"
        if "PC" not in self.table.columns:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"
        if dataincommentrow:
            string += self.swcomment
            string += self.krwcomment
            string += self.krowcomment
            if not self.fast:
                string += f"-- krw = krow @ sw={self.crosspoint():1.5f}\n"
            string += self.pccomment
        width = 10
        string += (
            "-- "
            + "SW".ljust(width - 3)
            + "KRW".ljust(width)
            + "KROW".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.table[["SW", "KRW", "KROW", "PC"]],
            monotonicity={
                "KROW": {"sign": -1, "lower": 0, "upper": 1},
                "KRW": {"sign": 1, "lower": 0, "upper": 1},
                "PC": {"sign": -1, "allowzero": True},
            }
            if not self.fast
            else None,
        )
        string += "/\n"  # Empty line at the end
        return string

    def SWFN(
        self,
        header: bool = True,
        dataincommentrow: bool = True,
        swcomment: Optional[str] = None,
        crosspointcomment: Optional[str] = None,
    ) -> str:
        """Return a SWFN keyword with data to Eclipse

        The columns sw, krw and pc are outputted and formatted
        accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SWFN string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SWFN yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
            swcomment: String to be used for swcomment, overrides what
                this object can provide. Used by GasWater
            crosspointcomment: String to be used for crosspoint comment
                string, overrides what this object can provide. Used by GasWater.
                If None, it will be computed, use empty string to avoid.
        """
        if not self.selfcheck(mode="SWFN"):
            # selfcheck will print errors/warnings
            return ""
        string = ""
        if "PC" not in self.table.columns:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SWFN\n"
        string += comment_formatter(self.tag)
        string += "-- pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            if swcomment is not None:
                string += swcomment
            else:
                string += self.swcomment
            string += self.krwcomment
            if crosspointcomment is None:
                if "KROW" in self.table.columns and not self.fast:
                    string += f"-- krw = krow @ sw={self.crosspoint():1.5f}\n"
            else:
                string += crosspointcomment
            string += self.pccomment
        width = 10
        string += (
            "-- "
            + "SW".ljust(width - 3)
            + "KRW".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.table[["SW", "KRW", "PC"]],
            monotonicity={
                "KRW": {"sign": 1, "lower": 0, "upper": 1},
                "PC": {"sign": -1, "allowzero": True},
            }
            if not self.fast
            else None,
        )
        string += "/\n"  # Empty line at the end
        return string

    def WOTABLE(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a string for a Nexus WOTABLE"""
        string = ""
        if "PC" not in self.table.columns:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"

        if header:
            string += "WOTABLE\n"
            string += "SW KRW KROW PC\n"
        string += "! pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            string += self.swcomment.replace("--", "!")
            string += self.krwcomment.replace("--", "!")
            string += self.krowcomment.replace("--", "!")
            if not self.fast:
                string += f"! krw = krow @ sw={self.crosspoint():1.5f}\n"
            string += self.pccomment.replace("--", "!")
        width = 10
        string += (
            "! "
            + "SW".ljust(width - 2)
            + "KRW".ljust(width)
            + "KROW".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.table[["SW", "KRW", "KROW", "PC"]],
            monotonicity={
                "KROW": {"sign": -1, "lower": 0, "upper": 1},
                "KRW": {"sign": 1, "lower": 0, "upper": 1},
                "PC": {"sign": -1, "allowzero": True},
            }
            if not self.fast
            else None,
        )
        return string

    def plotpc(
        self,
        mpl_ax=None,
        color: str = "blue",
        alpha: float = 1,
        linewidth: int = 1,
        linestyle: str = "-",
        label: str = "",
        logyscale: bool = False,
    ) -> None:
        """Plot capillary pressure (pc)

        If mpl_ax is supplied, the curve will be drawn on
        that, if not, a new axis (plot) will be made
        """
        # Lazy import for speed reasons.
        import matplotlib  # noqa: PLC0415
        import matplotlib.pyplot as plt  # noqa: PLC0415

        if mpl_ax is None:
            matplotlib.style.use("ggplot")
            _, useax = plt.subplots()
        else:
            useax = mpl_ax
        if logyscale:
            useax.set_yscale("log")
            useax.set_ylim((1e-6, 100))
        self.table.plot(
            ax=useax,
            x="SW",
            y="PC",
            c=color,
            alpha=alpha,
            label=label,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        if mpl_ax is None:
            plt.show()

    def plotkrwkrow(
        self,
        mpl_ax=None,
        color: str = "blue",
        alpha: float = 1,
        linewidth: float = 1,
        linestyle: str = "-",
        marker: Optional[str] = None,
        label: str = "",
        logyscale: bool = False,
    ) -> None:
        """Plot krw and krow

        If the argument 'mpl_ax' is not supplied, a new plot
        window will be made. If supplied, it will draw on
        the specified axis."""
        # Lazy import for speed reaons.
        import matplotlib  # noqa: PLC0415
        import matplotlib.pyplot as plt  # noqa: PLC0415

        if mpl_ax is None:
            matplotlib.style.use("ggplot")
            _, useax = plt.subplots()
        else:
            useax = mpl_ax
        if logyscale:
            useax.set_yscale("log")
            useax.set_ylim((1e-8, 1))
        self.table.plot(
            ax=useax,
            x="SW",
            y="KRW",
            c=color,
            alpha=alpha,
            legend=None,
            label=label,
            linewidth=linewidth,
            linestyle=linestyle,
            marker=marker,
        )
        self.table.plot(
            ax=useax,
            x="SW",
            y="KROW",
            c=color,
            alpha=alpha,
            label=None,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
            marker=marker,
        )
        if mpl_ax is None:
            plt.show()



"""Test the plotting module"""

from pathlib import Path

import matplotlib.pyplot as plt
import pytest

from pyscal import GasOil, GasWater, PyscalList, WaterOil, WaterOilGas, plotting


def test_get_satnum_from_tag():
    """Check that the SATNUM number can be retrieved from the model tag"""
    # Several PyscalLists of different model types to be checked
    pyscal_lists = [
        PyscalList(
            [
                WaterOil(tag="SATNUM 1"),
                WaterOil(tag="SATNUM 2"),
                WaterOil(tag="SATNUM 3"),
            ]
        ),
        PyscalList(
            [
                GasWater(tag="SATNUM 1"),
                GasWater(tag="SATNUM 2"),
                GasWater(tag="SATNUM 3"),
            ]
        ),
        PyscalList(
            [GasOil(tag="SATNUM 1"), GasOil(tag="SATNUM 2"), GasOil(tag="SATNUM 3")]
        ),
        PyscalList(
            [
                WaterOilGas(tag="SATNUM 1"),
                WaterOilGas(tag="SATNUM 2"),
                WaterOilGas(tag="SATNUM 3"),
            ]
        ),
    ]

    # Test that an integer is returned from this function
    for pyscal_list in pyscal_lists:
        for model in pyscal_list.pyscal_list:
            satnum = plotting.get_satnum_from_tag(model.tag)
            assert isinstance(satnum, int)

    # Test that this function doesn't work if SATNUM is represented as a float
    # in the model.tag string
    with pytest.raises(ValueError):
        satnum = plotting.get_satnum_from_tag(WaterOil(tag="SATNUM 3.0").tag)


def test_plotter():
    """Check if an Exception is raised if a model type is not included. This is done
    to check that all models have been implemented in the plotting module."""

    class DummyPyscalList:
        """
        Can't use the actual PyscalList, as this will raise its own exception
        (DummyModel is not a pyscal object), so a dummy PyscalList is used

        #If the PyscalList.pyscal_list instance variable name changes, this
        will still pass..."""

        def __init__(self, models: list) -> None:
            self.pyscal_list = models

    class DummyModel:
        """Dummy model"""

        def __init__(self, tag: str) -> None:
            self.tag = tag

    dummy_model_list = [
        DummyModel("SATNUM 1"),
        DummyModel("SATNUM 2"),
        DummyModel("SATNUM 3"),
    ]

    dummy_pyscal_list = DummyPyscalList(dummy_model_list)

    with pytest.raises(Exception):
        plotting.plotter(dummy_pyscal_list)


def test_pyscal_list_attr():
    """Check that the PyscalList class has an pyscal_list instance variable.
    This is accessed by the plotting module to loop through models to plot."""
    assert hasattr(PyscalList(), "pyscal_list") is True, (
        "The PyscalList object should have a pyscal_list instance variable.\
        This is accessed by the plotting module."
    )


def test_plot_relperm():
    """Test that a matplotlib.pyplot Figure instance is returned"""
    wateroil = WaterOil(swl=0.1, h=0.1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()

    kwargs = {"semilog": False, "pc": False}
    config = plotting.get_plot_config_options("WaterOil", **kwargs)
    fig = plotting.plot_relperm(wateroil.table, 1, config, **kwargs)

    assert isinstance(fig, plt.Figure), (
        "Type of returned object is not a matplotlib.pyplot Figure"
    )


def test_plot_pc():
    """Test that a matplotlib.pyplot Figure instance is returned"""
    wateroil = WaterOil(swl=0.1, h=0.1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    wateroil.add_simple_J()

    kwargs = {"semilog": False, "pc": True}
    config = plotting.get_plot_config_options("WaterOil", **kwargs)
    fig = plotting.plot_pc(wateroil.table, 1, **config)

    assert isinstance(fig, plt.Figure), (
        "Type of returned object is not a matplotlib.pyplot Figure"
    )


def test_wog_plotter(tmpdir):
    """Test that relative permeability figures are created by the plotter function"""
    wateroil = WaterOil(swl=0.1, h=0.1, tag="SATNUM 1")
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    wateroil.add_simple_J()

    gasoil = GasOil(swl=0.1, h=0.1, tag="SATNUM 1")
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()

    wateroilgas = WaterOilGas()
    wateroilgas.wateroil = wateroil
    wateroilgas.gasoil = gasoil

    # Testing both rel perm and Pc plots are created
    kwargs = {"semilog": False, "pc": True, "outdir": tmpdir}
    kr_ow_plot_name = "krw_krow_SATNUM_1.png"
    kr_og_plot_name = "krw_krow_SATNUM_1.png"
    pc_ow_plot_name = "pcow_SATNUM_1.png"

    plotting.wog_plotter(wateroilgas, **kwargs)

    assert Path.exists(Path(tmpdir).joinpath(kr_ow_plot_name)), "Plot not created"
    assert Path.exists(Path(tmpdir).joinpath(kr_og_plot_name)), "Plot not created"
    assert Path.exists(Path(tmpdir).joinpath(pc_ow_plot_name)), "Plot not created"


def test_wo_plotter(tmpdir):
    """Test that relative permeability figures are created by the plotter function"""
    wateroil = WaterOil(swl=0.1, h=0.1, tag="SATNUM 1")
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    wateroil.add_simple_J()

    # Testing both rel perm and Pc plots are created
    kwargs = {"semilog": False, "pc": True, "outdir": tmpdir}
    kr_plot_name = "krw_krow_SATNUM_1.png"
    pc_plot_name = "pcow_SATNUM_1.png"

    plotting.wo_plotter(wateroil, **kwargs)

    assert Path.exists(Path(tmpdir).joinpath(kr_plot_name)), "Plot not created"
    assert Path.exists(Path(tmpdir).joinpath(pc_plot_name)), "Plot not created"


def test_wo_plotter_relperm_only(tmpdir):
    """Test that relative permeability figures are created by the plotter function"""
    wateroil = WaterOil(swl=0.1, h=0.1, tag="SATNUM 1")
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    wateroil.add_simple_J()

    # Testing only relperm plots are created
    kwargs = {"semilog": False, "pc": False, "outdir": tmpdir}
    kr_plot_name = "krw_krow_SATNUM_1.png"
    pc_plot_name = "pcow_SATNUM_1.png"

    plotting.wo_plotter(wateroil, **kwargs)

    assert Path.exists(Path(tmpdir).joinpath(kr_plot_name)), "Plot not created"
    assert not Path.exists(Path(tmpdir).joinpath(pc_plot_name)), (
        "Plot created when it shouldn't have been"
    )


def test_go_plotter(tmpdir):
    """Test that relative permeability figures are created by the plotter function"""
    gasoil = GasOil(swl=0.1, h=0.1, tag="SATNUM 1")
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()

    # Add Pc manually as there are currently no supporting functions for adding
    # Pc to GasOil instances
    gasoil.table["PC"] = 1 * len(gasoil.table)

    # Testing both rel perm and Pc plots are created
    kwargs = {"semilog": False, "pc": True, "outdir": tmpdir}
    kr_plot_name = "krg_krog_SATNUM_1.png"
    pc_plot_name = "pcog_SATNUM_1.png"

    plotting.go_plotter(gasoil, **kwargs)

    assert Path.exists(Path(tmpdir).joinpath(kr_plot_name)), "Plot not created"
    assert Path.exists(Path(tmpdir).joinpath(pc_plot_name)), "Plot not created"


def test_gw_plotter(tmpdir):
    """Test that relative permeability figures are created by the plotter function"""
    gaswater = GasWater(swl=0.1, h=0.1, tag="SATNUM 1")
    gaswater.add_corey_water()
    gaswater.add_corey_gas()
    gaswater.add_simple_J()

    # Testing both rel perm and Pc plots are created
    kwargs = {"semilog": False, "pc": True, "outdir": tmpdir}
    kr_plot_name = "krg_krw_SATNUM_1.png"
    pc_plot_name = "pcgw_SATNUM_1.png"

    plotting.gw_plotter(gaswater, **kwargs)

    assert Path.exists(Path(tmpdir).joinpath(kr_plot_name)), "Plot not created"
    assert Path.exists(Path(tmpdir).joinpath(pc_plot_name)), "Plot not created"


def test_save_figure(tmpdir):
    """Test that figure is saved"""
    fig = plt.Figure()

    config = {"curves": "dummy", "suffix": ""}
    fig_name = "dummy_SATNUM_1.png"
    plotting.save_figure(fig, 1, config, plot_type="relperm", outdir=tmpdir)

    assert Path.exists(Path(tmpdir).joinpath(fig_name)), "Figure not saved"



"""Common functions and mock data for usage in pyscal testing"""

import io
from contextlib import suppress
from typing import Union

import numpy as np
import pandas as pd

with suppress(ModuleNotFoundError):
    # This module is only needed for testing, but this testing module
    # should be importable in other scenarios as well.
    from hypothesis import settings

from pyscal import GasOil, WaterOil


def slow_hypothesis(slow_function):
    """A decorator for tests that are always slow.

    The default deadline for hypothesis is set through the hypothesis profile
    being used, and is set to None for CI runs (through a command line argument
    and code in conftest.py), meaning no deadline. On tests that are slow on
    local iron as well, we must ensure that the deadline set for these do not
    override the "no-deadline" in CI.
    """
    if settings().deadline is None:
        return settings()(slow_function)
    return settings(deadline=1000)(slow_function)


def series_decreasing(series: pd.Series):
    """Weaker than pd.Series.is_monotonic_decreasing,
    allows constant parts.

    We do not enforce less than zero here, because there will be
    some positive differences due to representation errors which
    is ok in terms of numerical equivalence. It is not ok when dumped
    to Eclipse, so the representation of the dataframe as a SGOF table
    must be properly rounded before printed."""
    return (series.diff().dropna() < 1e-8).all()


def series_increasing(series: pd.Series):
    """Weaker than pd.Series.is_monotonic_increasing"""
    return (series.diff().dropna() > -1e-8).all()


def sat_table_str_ok(sat_table_str: str) -> None:
    """Test that a supplied string from SWOF()/SGOF() etc is
    probably ok for Eclipse.

    Number of floats pr. line must be constant
    All numerical lines must be parseable to a rectangular dataframe
    with only floats. The first column must contain only unique values
    for every SATNUM.
    """
    assert sat_table_str

    for line in sat_table_str.splitlines():
        try:
            if not (not line or line.startswith(("S", "--", "/")) or int(line[0]) >= 0):
                assert False

        except ValueError as e_msg:
            # the int(line[0]) will get here on strings.
            print(e_msg)
            assert False

    assert "-- pyscal: " in sat_table_str

    # On non-comment lines, number of ascii floats should be the same:
    number_lines = [
        line
        for line in sat_table_str.splitlines()
        if line.strip() and line.strip()[0] in ["0", "1", "."]
    ]

    floats_pr_line = {len(line.split()) for line in number_lines}
    # This must be a constant:
    assert len(floats_pr_line) == 1
    # And not more than 4:
    if not next(iter(floats_pr_line)) <= 4:
        print(sat_table_str)
    assert next(iter(floats_pr_line)) <= 4

    float_characters = {len(flt) for flt in " ".join(number_lines).split()}
    digits = 7  # This is the default value in utils.df2str()
    for float_str_length in float_characters:
        assert not 1 < float_str_length < digits + 2
        # float_str_length must be 1 (a pure zero value),
        # or above digits + 1, otherwise it is a sign of some error.

    # And pyscal only emits three or four floats pr. line for all keywords:
    assert next(iter(set(floats_pr_line))) in [3, 4]

    # So we should be able to parse this to a dataframe:
    dframe = pd.read_csv(io.StringIO("\n".join(number_lines)), sep=" ", header=None)
    assert len(dframe) == len(number_lines)

    # The first column holds saturations, for pyscal test-data that
    # is always between zero and 1
    assert 0 <= dframe[0].min() <= dframe[0].max() <= 1

    # Saturations should be unique, but only within each SATNUM.
    # Assert this by checking that the two consecutive numbers in the
    # first column are never the same:
    assert (~np.isclose(dframe[0].diff().dropna(), 0)).all()

    # Second column is never capillary pressure, so there we can enforce the same
    assert 0 <= dframe[1].min() <= dframe[1].max() <= 1
    # And then sometimes for the third column:
    if len(dframe.columns) > 3 or "SOF3" in sat_table_str:
        assert 0 <= dframe[2].min() <= dframe[2].max() <= 1


def check_table(dframe: pd.DataFrame) -> None:
    """Check that the numbers in a dataframe for WaterOil or GasOil
    has the properties that Eclipse enforces"""
    assert not dframe.empty
    assert not dframe.isna().to_numpy().any()
    if "SW" in dframe and "SG" not in dframe:
        # (avoiding GasWater tables, where sw is an auxiliary column)
        assert len(dframe["SW"].unique()) == len(dframe)
        assert dframe["SW"].is_monotonic_increasing
        assert (dframe["SW"] >= 0.0).all()
        assert dframe["SWN"].is_monotonic_increasing
        assert dframe["SON"].is_monotonic_decreasing
        assert dframe["SWNPC"].is_monotonic_increasing
    if "SG" in dframe:
        assert len(dframe["SG"].unique()) == len(dframe)
        assert dframe["SG"].is_monotonic_increasing
        assert (dframe["SG"] >= 0.0).all()
        assert dframe["SGN"].is_monotonic_increasing
        assert dframe["SON"].is_monotonic_decreasing
    if "KROW" in dframe:
        assert series_decreasing(dframe["KROW"])
        assert (dframe["KROW"] >= 0).all()
        assert (dframe["KROW"] <= 1.0).all()
    if "KRW" in dframe:
        assert series_increasing(dframe["KRW"])
        assert np.isclose(dframe["KRW"].iloc[0], 0.0)
        assert (dframe["KRW"] >= 0).all()
        assert (dframe["KRW"] <= 1.0).all()
    if "PC" in dframe:
        if "SW" in dframe:
            assert series_decreasing(dframe["PC"])
        if "SG" in dframe:
            assert series_increasing(dframe["PC"])
    if "KROG" in dframe:
        assert series_decreasing(dframe["KROG"])
        assert (dframe["KROG"] >= 0).all()
        assert (dframe["KROG"] <= 1.0).all()
    if "KRG" in dframe:
        assert series_increasing(dframe["KRG"])
        assert (dframe["KRG"] >= 0).all()
        assert (dframe["KRG"] <= 1.0).all()


def check_linear_sections(wo_or_go: Union[WaterOil, GasOil]) -> None:
    """Check that the linear segments of a WaterOil or a GasOil
    object are linear."""
    if isinstance(wo_or_go, WaterOil):
        sat_col = "SW"
        right_start = 1.0 - wo_or_go.sorw
        right_end = 1.0
        right_lin_cols = ["KROW", "KRW"]
        left_lin_cols = ["KRW"]
        left_start = wo_or_go.swl
        left_end = wo_or_go.swcr
    if isinstance(wo_or_go, GasOil):
        sat_col = "SG"
        if wo_or_go.krgendanchor == "sorg":
            right_start = 1.0 - wo_or_go.sorg - wo_or_go.swl
        else:
            # If not krgendanchor=sorg, then there is no linear
            # segment to the right.
            right_start = 1.0 - wo_or_go.swl
        right_end = 1.0 - wo_or_go.swl
        right_lin_cols = ["KROG", "KRG"]
        left_lin_cols = ["KRG"]
        left_start = 0.0
        left_end = wo_or_go.sgcr

    right_lin_seg = wo_or_go.table[
        (wo_or_go.table[sat_col] >= right_start)
        & (wo_or_go.table[sat_col] <= right_end)
    ]
    left_lin_seg = wo_or_go.table[
        (wo_or_go.table[sat_col] >= left_start) & (wo_or_go.table[sat_col] <= left_end)
    ]
    if len(right_lin_seg) > 5:
        for col in right_lin_cols:
            # We avoid the first and two lasts row in right_lin_seg, because
            # this does not always match the constant saturation segment
            # assumption in this linearity test:
            assert right_lin_seg.iloc[1:-2][col].diff().std() < 1e-9
    if len(left_lin_seg) > 4:
        for col in left_lin_cols:
            assert left_lin_seg.iloc[1:-1][col].diff().std() < 1e-9


def float_df_checker(
    dframe: pd.DataFrame, idxcol: str, value: float, compcol: str, answer: float
) -> Union[np.bool_, np.ndarray]:
    """Looks up in a dataframe, selects the row where idxcol=value
    and compares the value in compcol with answer

    Warning: This is slow code, but only the tests are slow

    Floats are notoriously difficult to handle in computers.
    """
    # Find row index where we should do comparison:
    plus_one = 0
    if abs(answer) < 0.2:
        plus_one = 1
    for swtol in [1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1]:
        idxs = (dframe[idxcol] - value).abs() < swtol
        if sum(idxs) < 1:
            continue
        if sum(idxs) < 10:
            break
    rowidx = (dframe[idxs][idxcol] - value).abs().sort_values().index[0]
    return np.isclose(plus_one + dframe.loc[rowidx, compcol], plus_one + answer)
